from __future__ import annotations

import numpy as np
import xarray as xr
from scipy.integrate import cumulative_trapezoid

import sasktran2 as sk


class SolarModel:
    def __init__(
        self,
        source="solar_irradiance_hsrs_2022_11_30_extended",
        ds=None,
        mode="sample",
        resolution=None,
        resolution_in_wavelength=True,
    ):
        """
        Loads in a solar model for use inside SASKTRAN2.  By default the HSRS extended solar model is loaded in
        that has a useful wavelength range covering 115 nm to 200 microns and intgrates to 1362.8 W/m^2.

        Coddington, O. M., Richard, E. C., Harber, D., Pilewskie, P., Woods, T. N., Snow, M., et al. (2023). Version 2 of the TSIS-1 Hybrid Solar Reference Spectrum and Extension to the Full Spectrum. Earth and Space Science, 10, e2022EA002637. https://doi.org/10.1029/2022EA002637

        Optionally a manually specified dataset can be used, which must contain the variables "wavelength" and "irradiance". It is assumed
        that "wavelength" is in units of nm and "irradiance" has spectral units of /nm.

        Parameters
        ----------
        source: str, optional
            sasktran2 database source identifier.  Currently only "solar_irradiance_hsrs_2022_11_30_extended" is supported and is the default
        ds: xr.Dataset, optional
            If set, then the dataset is used instead of the source.  The dataset must contain the variables "wavelength" and "irradiance"
            default is None
        mode: str, optional
            The mode of the solar model.  Options are "integrate", "average", "sample".  Default is "sample".  If "integrate" is selected
            then the solar spectrum is integrated over the wavelength range of interest.  If "average" is selected then the solar spectrum
            is averaged over the wavelength range of interest.  If "sample" is selected then the solar spectrum is sampled at the wavelengths
            of interest.  The "integrate" and "average" modes are useful for calculating the solar irradiance at a specific wavelength range
            and the "sample" mode is useful for calculating the solar irradiance at specific wavelengths.
        resolution: float, optional
            The resolution to integrate the solar spectrum at.  If None, then the resolution is determined by the wavelength grid.  Default is None
        resolution_in_wavelength: bool, optional
            If True, then the resolution is in wavelength units.  If False, then the resolution is in wavenumber units.  Default is True
        """
        if ds is not None:
            self._ds = ds
        else:
            self._ds = sk.database.StandardDatabase().path(f"solar/{source}.nc")
            self._ds = xr.open_dataset(self._ds)

        self._wv = self._ds["wavelength"].to_numpy()
        self._irrad = self._ds["irradiance"].to_numpy()
        self._ds.close()

        self._mode = mode.lower()

        if self._mode == "integrate" or self._mode == "average":
            self._integral = cumulative_trapezoid(
                self._ds["irradiance"].values, self._ds["wavelength"].values
            )
        elif self._mode == "sample":
            self._integral = None
        else:
            msg = "Invalid mode"
            raise ValueError(msg)

        self._resolution = resolution
        self._resolution_in_wavelength = resolution_in_wavelength

    def irradiance(
        self, wavelengths: np.ndarray, solardistance: float | None = None
    ) -> np.array:
        """
        Calculates the solar irradiance at the specified wavelengths in nm.  Default units
        are in W/m^2/nm, but can be different if a manual database is used instead.  If the mode is set to "integrate"
        then the units will be W/m^2.

        The solar distance factor can be used to scale the solar irradiance to a different distance.  The default is None
        indicating that the solar irradiance is at 1 AU.

        Parameters
        ----------
        wavelengths : np.ndarray
            Wavelengths in [nm] to calculate the solar irradiance at
        solardistance : float, optional
            Solar distance to scale by in AU., by default None

        Returns
        -------
        np.array
            The solar irradiance at the specified wavelengths
        """
        solar_distance_factor = (
            1 / (solardistance**2) if solardistance is not None else 1
        )

        if self._mode == "sample":
            return np.interp(wavelengths, self._wv, self._irrad) * solar_distance_factor
        # Have to calculate the wavelength intervals
        # and integrate the spectrum over the intervals

        if self._resolution is not None:
            if self._resolution_in_wavelength:
                left_intervals = wavelengths - self._resolution / 2
                right_intervals = wavelengths + self._resolution / 2
            else:
                left_intervals = 1e7 / (1e7 / wavelengths + self._resolution / 2)
                right_intervals = 1e7 / (1e7 / wavelengths - self._resolution / 2)
        else:
            left_intervals = np.concat(
                ([wavelengths[0]], np.diff(wavelengths) / 2 + wavelengths[:-1])
            )
            right_intervals = np.concat(
                (np.diff(wavelengths) / 2 + wavelengths[:-1], [wavelengths[-1]])
            )

        lidx = np.searchsorted(self._ds["wavelength"].values, left_intervals) - 1
        ridx = np.searchsorted(self._ds["wavelength"].values, right_intervals) - 1

        left_vals = np.interp(left_intervals, self._wv, self._irrad)
        right_vals = np.interp(right_intervals, self._wv, self._irrad)

        # Cumlative integral to left side
        I_left = self._integral[lidx] + 0.5 * (self._irrad[lidx] + left_vals) * (
            left_intervals - self._wv[lidx]
        )
        I_right = self._integral[ridx] + 0.5 * (self._irrad[ridx] + right_vals) * (
            right_intervals - self._wv[ridx]
        )

        if self._mode == "integrate":
            return (I_right - I_left) * solar_distance_factor
        return (
            (I_right - I_left)
            / (right_intervals - left_intervals)
            * solar_distance_factor
        )

        # TODO: Implement doppler shift?



from __future__ import annotations

import numpy as np
import pytest
import sasktran2 as sk


def test_hitran_online():
    pytest.importorskip("hapi")

    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    wavel = np.arange(1360, 1370, 0.01)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geometry.altitudes(),
        np.ones_like(model_geometry.altitudes()) * 1e-6,
    )

    atmosphere["h2o"] = sk.climatology.mipas.constituent(
        "H2O", sk.optical.HITRANAbsorber("H2O")
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)

    _ = engine.calculate_radiance(atmosphere)



from __future__ import annotations

import numpy as np
import pandas as pd
import xarray as xr
from scipy import interpolate

from sasktran2._core import (
    Geometry1D,
    GeometryType,
    InterpolationMethod,
    TangentAltitudeSolar,
    ViewingGeometry,
)
from sasktran2.geodetic import WGS84
from sasktran2.solar import SolarGeometryHandlerBase

from .base import ViewingGeometryContainer


class LimbVertical(ViewingGeometry, ViewingGeometryContainer):
    def __init__(
        self,
        solar_handler: SolarGeometryHandlerBase,
        tangent_altitudes: np.array,
        tangent_latitude: float | np.ndarray,
        tangent_longitude: float | np.ndarray,
        time: pd.Timestamp | np.ndarray,
        observer_altitude: float | np.ndarray,
        observer_latitude: float | np.ndarray,
        observer_longitude: float | np.ndarray,
        reference_altitude: float = 25000,
    ):
        """
        A Limb Vertical image is a geometry container where the observer is looking through the limb of the atmosphere, at a set
        of tangent altitudes.

        We recommend constructing this class through one of the various `from_` class methods.  No validation checking
        is done through this constructor.

        Parameters
        ----------
        solar_handler : sk.solar.SolarGeometryHandlerBase
            Solar geometry handler to use for calculating solar angles.
        tangent_altitudes : np.array
            Altitudes of the tangent points.
        tangent_latitude : float | np.ndarray
            Latitude of the tangent points. Can be a scalar or same size as `tangent_altitudes`.
        tangent_longitude : float | np.ndarray
            Longitude of the tangent points. Can be a scalar or same size as `tangent_altitudes`.
        time : pd.Timestamp | np.ndarray
            Time of the observation. Can be a scalar or same size as `tangent_altitudes`.
        observer_altitude : float | np.ndarray
            Altitude of the observer. Can be a scalar or same size as `tangent_altitudes`.
        observer_latitude : float | np.ndarray
            Latitude of the observer. Can be a scalar or same size as `tangent_altitudes`.
        observer_longitude : float | np.ndarray
            Longitude of the observer. Can be a scalar or same size as `tangent_altitudes`.
        reference_altitude : float, optional
            The tangent altitude where reference parameters such as earth radius
            are defined at, by default 25000.
        """
        self._tangent_altitudes = np.array(tangent_altitudes)

        # Determine the expected length
        n = len(self._tangent_altitudes)

        # Helper function to convert scalars to arrays
        def to_array(param):
            if np.isscalar(param) | isinstance(param, pd.Timestamp):
                return np.full(n, param)
            param = np.array(param)
            if len(param) != n:
                msg = "Parameter length mismatch."
                raise ValueError(msg)
            return param

        self._tangent_latitude = to_array(tangent_latitude)
        self._tangent_longitude = to_array(tangent_longitude)
        self._time = to_array(time)
        self._observer_altitude = to_array(observer_altitude)
        self._observer_latitude = to_array(observer_latitude)
        self._observer_longitude = to_array(observer_longitude)

        self._reference_altitude = reference_altitude

        ViewingGeometry.__init__(self)

        tangent_geo = WGS84()
        observer_geo = WGS84()

        viewing_zenith = np.zeros(n)
        viewing_azimuth = np.zeros(n)

        self._cos_sza = np.zeros(n)
        self._earth_radius = np.zeros(n)
        self._solar_azimuth = np.zeros(n)
        self._observer_azimuth = np.zeros(n)

        for i, (alt, lat, lon, t, obs_alt, obs_lat, obs_lon) in enumerate(
            zip(
                tangent_altitudes,
                self._tangent_latitude,
                self._tangent_longitude,
                self._time,
                self._observer_altitude,
                self._observer_latitude,
                self._observer_longitude,
                strict=False,
            )
        ):
            tangent_geo.from_lat_lon_alt(lat, lon, alt)
            observer_geo.from_lat_lon_alt(obs_lat, obs_lon, obs_alt)

            solar_zenith, solar_azimuth = solar_handler.target_solar_angles(
                lat, lon, alt, t
            )

            # Need to calculate the relative azimuth
            lv = tangent_geo.location - observer_geo.location
            lv /= np.linalg.norm(lv)

            # Need angle in the (north, west) plane measured clockwise
            observer_azimuth = -np.arctan2(
                np.dot(lv, tangent_geo.local_west), -np.dot(lv, tangent_geo.local_south)
            )

            # Now observer_azimuth is pointing away from the observer, and solar azimuth is pointing towards the sun
            # If they are equal this is forward scatter as expected,

            self.add_ray(
                TangentAltitudeSolar(
                    tangent_altitude_m=alt,
                    relative_azimuth=np.deg2rad(solar_azimuth - observer_azimuth),
                    observer_altitude_m=obs_alt,
                    cos_sza=np.cos(np.deg2rad(solar_zenith)),
                )
            )

            self._observer_azimuth[i] = observer_azimuth
            self._solar_azimuth[i] = solar_azimuth

            # Assign the viewing zenith and azimuthal angles
            viewing_zenith[i] = np.rad2deg(np.arccos(np.dot(lv, observer_geo.local_up)))

            viewing_azimuth[i] = -np.arctan2(
                np.dot(lv, observer_geo.local_west),
                -np.dot(lv, observer_geo.local_south),
            )

            # Store parameters
            self._cos_sza[i] = np.cos(np.deg2rad(solar_zenith))
            self._earth_radius[i] = np.linalg.norm(
                tangent_geo.location - alt * tangent_geo.local_up
            )

        geometry_ds = xr.Dataset(
            {
                "tangent_altitude": (["los"], tangent_altitudes),
                "tangent_longitude": (
                    ["los"],
                    np.ones_like(tangent_altitudes) * tangent_longitude,
                ),
                "tangent_latitude": (
                    ["los"],
                    np.ones_like(tangent_altitudes) * tangent_latitude,
                ),
                "time": (["los"], self._time.astype(np.datetime64)),
                "observer_altitude": (["los"], self._observer_altitude),
                "observer_latitude": (["los"], self._observer_latitude),
                "observer_longitude": (["los"], self._observer_longitude),
                "tangent_cos_sza": (["los"], self._cos_sza),
                "tangent_solar_azimuth": (["los"], self._solar_azimuth),
                "tangent_observer_azimuth": (["los"], self._observer_azimuth),
                "viewing_zenith": (["los"], viewing_zenith),
                "viewing_azimuth": (["los"], viewing_azimuth),
            },
        )

        ViewingGeometryContainer.__init__(self, geometry_ds)

    def recommended_cos_sza(self) -> float:
        """
        Returns the cosine of the solar zenith angle at the reference altitude.

        Returns
        -------
        float
        """
        return interpolate.interp1d(self._tangent_altitudes, self._cos_sza)(
            self._reference_altitude
        )

    def recommended_earth_radius(self) -> float:
        """
        Returns the earth radius at the reference altitude.

        Returns
        -------
        float
        """
        return interpolate.interp1d(self._tangent_altitudes, self._earth_radius)(
            self._reference_altitude
        )

    def model_geometry(self, altitude_grid_m: np.ndarray) -> Geometry1D:
        return Geometry1D(
            self.recommended_cos_sza(),
            0.0,
            self.recommended_earth_radius(),
            altitude_grid_m,
            InterpolationMethod.LinearInterpolation,
            GeometryType.Spherical,
        )

    @classmethod
    def from_tangent_parameters(
        cls,
        solar_handler: SolarGeometryHandlerBase,
        tangent_altitudes: np.ndarray,
        tangent_latitude: float,
        tangent_longitude: float,
        time: pd.Timestamp,
        observer_altitude: float,
        viewing_azimuth: float,
        reference_altitude: float = 25000,
        forced_constant_tangent: bool = False,
    ):
        """
        Initialize a LimbVertical object from a set of tangent parameters assuming a vertical image is taken. From
        a single observer location.

        Parameters
        ----------
        solar_handler : sk.solar.SolarGeometryHandlerBase
            Solar geometry handler to use for calculating solar angles.
        tangent_altitudes : np.ndarray
            The altitudes of the tangent points.
        tangent_latitude : float
            Latitude of the tangent point at reference_altitude
        tangent_longitude : float
            Longitude of the tangent piont at reference_altitude
        time : pd.Timestamp
            Time of the observation
        observer_altitude : float
            Altitude of the observer
        viewing_azimuth : float
            Azimuth of the viewing direction in degrees clockwise from north at the tangent point.
        reference_altitude : float, optional
            Altitude the tangent point parameters are defined at, by default 25000
        forced_constant_tangent : bool, optional
            If true, then variation in latitude/longitude is not taken into account as a function of
            tangent altitude, by default False
        """
        tangent_geo = WGS84()

        tangent_geo.from_lat_lon_alt(
            tangent_latitude, tangent_longitude, reference_altitude
        )

        lv = -tangent_geo.local_south * np.cos(
            np.deg2rad(viewing_azimuth)
        ) - tangent_geo.local_west * np.sin(np.deg2rad(viewing_azimuth))

        # Now we have to find the observer location
        observer_geo = WGS84()
        observer_geo.from_xyz(
            observer_geo.altitude_intercepts(
                observer_altitude, tangent_geo.location, lv
            )[0]
        )

        # Then we can calculate all of the actual tangent latitudes and longitudes
        actual_tangent_latitude = np.zeros(len(tangent_altitudes))
        actual_tangent_longitude = np.zeros(len(tangent_altitudes))

        if forced_constant_tangent:
            actual_tangent_latitude = tangent_latitude
            actual_tangent_longitude = tangent_longitude
        else:
            for i, alt in enumerate(tangent_altitudes):
                tangent_geo.from_tangent_altitude(alt, observer_geo.location, lv)
                actual_tangent_latitude[i] = tangent_geo.latitude
                actual_tangent_longitude[i] = tangent_geo.longitude

        return cls(
            solar_handler,
            tangent_altitudes,
            actual_tangent_latitude,
            actual_tangent_longitude,
            time,
            observer_altitude,
            observer_geo.latitude,
            observer_geo.longitude,
            reference_altitude,
        )



from __future__ import annotations

import numpy as np

from sasktran2.constants import K_BOLTZMANN


class EquationOfState:
    def __init__(self, molar_mass_dry_air: float = 28.96e-3) -> None:
        """
        The equation of state relates pressure and temperature to number density and its derivatives.
        In addition we include specific humidity to convert between wet air and dry air

        Parameters
        ----------
        molar_mass_dry_air : float, optional
            Molar mass of dry air, by default 28.96e-3
        """
        self._pressure_pa = None
        self._temperature_k = None
        self._specific_humidity = None
        self._M = molar_mass_dry_air

    @property
    def pressure_pa(self) -> np.array:
        """
        Returns
        -------
        np.array
            Pressure in Pa
        """
        return self._pressure_pa

    @pressure_pa.setter
    def pressure_pa(self, pressure: np.array):
        self._pressure_pa = pressure

    @property
    def temperature_k(self) -> np.array:
        """

        Returns
        -------
        np.array
            Temperature in K
        """
        return self._temperature_k

    @temperature_k.setter
    def temperature_k(self, temperature: np.array):
        self._temperature_k = temperature

    @property
    def specific_humidity(self) -> np.array:
        """

        Returns
        -------
        np.array
            Specific humidity
        """
        return self._specific_humidity

    @specific_humidity.setter
    def specific_humidity(self, specific_humidity: np.array):
        self._specific_humidity = specific_humidity

    @property
    def air_numberdensity(self) -> np.array:
        """
        Converts pressure and temperature to number density using the ideal gas law at the grid sample points

        Returns
        -------
        dict with keys
            "N": Number density in [molecules / m^3]
            "dN_dP": Derivative of number density with respect to pressure in [molecules / m^3 / Pa]
            "dN_dT": Derivative of number density with respect to temperature in [molecules / m^3 / K]
        """
        # Ideal gas law is PV = N k T
        # Or (N/V) = P / (k T)

        return {
            "N": self.pressure_pa / (K_BOLTZMANN * self.temperature_k),
            "dN_dP": 1 / (K_BOLTZMANN * self.temperature_k),
            "dN_dT": -self.pressure_pa / (K_BOLTZMANN * self.temperature_k**2),
        }

    @property
    def dry_air_numberdensity(self) -> np.array:
        """
        Converts pressure and temperature to dry air number density using the ideal gas law at the grid sample points

        Returns
        -------
        dict with keys
            "N": Number density in [molecules / m^3]
            "dN_dP": Derivative of number density with respect to pressure in [molecules / m^3 / Pa]
            "dN_dT": Derivative of number density with respect to temperature in [molecules / m^3 / K]
            "dN_dsh": Derivative of number density with respect to specific humidity in [molecules / m^3 / kg/kg]
        """
        wet = self.air_numberdensity

        q = 0 if self.specific_humidity is None else self.specific_humidity

        return {
            "N": wet["N"] * (1 - q),
            "dN_dP": wet["dN_dP"] * (1 - q),
            "dN_dT": wet["dN_dT"] * (1 - q),
            "dN_dsh": -wet["N"],
        }



# ruff: noqa: F401
from __future__ import annotations

from .._core import LinearizedMie, MieData, MieOutput
from .distribution import LogNormalDistribution, integrate_mie
from .refractive import RefractiveIndex



from __future__ import annotations

import numpy as np

from sasktran2.atmosphere import Atmosphere
from sasktran2.util.interpolation import linear_interpolating_matrix

from ..._core import (  # noqa: F401
    LambertianStokes_1,
    LambertianStokes_3,
    MODISStokes_1,
    MODISStokes_3,
    SnowKokhanovskyStokes_1,
    SnowKokhanovskyStokes_3,
)


class WavelengthInterpolatorMixin:
    def __init__(
        self,
        wavelengths_nm: np.array = None,
        wavenumbers_cminv: np.array = None,
        out_of_bounds_mode="zero",
        param_length: int = 1,
    ):
        """
        A MixIn class that provides interpolation functionality over wavelength or wavenumber

        Parameters
        ----------
        wavelengths_nm : np.array, optional
            Wavelengths in [nm] that the parameters are specified at, by default None
        wavenumbers_cminv : np.array, optional
            Wavenumbers in inverse cm that the parameters are specified at, by default None
        out_of_bounds_mode : str, optional
            One of ["extend" or "zero"], "extend" will extend the last/first value if we are
            interpolating outside the grid. "zero" will set the parameters to 0 outside of the
            grid boundaries, by default "zero"
        """
        self._out_of_bounds_mode = out_of_bounds_mode

        if wavelengths_nm is not None:
            self._x = np.atleast_1d(wavelengths_nm)
            self._interp_var = "wavelengths_nm"
        elif wavenumbers_cminv is not None:
            self._x = np.atleast_1d(wavenumbers_cminv)
            self._interp_var = "wavenumbers_cminv"
        else:
            if param_length == 1:
                self._interp_var = "constant"
            else:
                self._interp_var = "native"

    def _interpolating_matrix(self, atmo: Atmosphere):
        if self._interp_var == "constant":
            # Don't have to interpolate, just assign the constant value
            return np.ones(atmo.num_wavel).reshape(-1, 1)
        if self._interp_var == "native":
            # Also can just assign the user value, but first make sure that it is the correct length
            if len(self._albedo) != atmo.num_wavel:
                msg = "The number of albedo values must match the number of wavelengths in the atmosphere"
                raise ValueError(msg)
            return np.eye(atmo.num_wavel)
        # Now we have to interpolate
        grid_x = getattr(atmo, self._interp_var)

        if grid_x is None:
            msg = f"The atmosphere does not have {self._interp_var} defined, cannot interpolate the user albedo"
            raise ValueError(msg)
        return linear_interpolating_matrix(self._x, grid_x, self._out_of_bounds_mode)



from __future__ import annotations

import numpy as np

from sasktran2.atmosphere import Atmosphere

from ..base import Constituent
from . import (
    MODISStokes_1,
    MODISStokes_3,
    WavelengthInterpolatorMixin,
)


class MODIS(Constituent, WavelengthInterpolatorMixin):
    def __init__(
        self,
        isotropic: np.array,
        volumetric: np.array = 0.0,
        geometric: np.array = 0.0,
        wavelengths_nm: np.array = None,
        out_of_bounds_mode="zero",
    ) -> None:
        """
        Parameters
        ----------
        isotropic : np.array
            Isotropic component (contribution weight of Lambertian).
        volumetric : np.array, optional
            Volumetric component (contribution weight of RossThick kernel), by default 0.
        geometric : np.array, optional
            Geometric component (contribution weight of LiSparse-R kernel), by default 0.
        wavelengths_nm : np.array, optional
            Wavelengths in [nm] that the parameters isotropic, volumetric, geometric is specified at, by default None indicating that these parameters are scalar.
        out_of_bounds_mode : str, optional
            One of ["extend" or "zero"], "extend" will extend the last/first value if we are
            interpolating outside the grid. "zero" will set the albedo to 0 outside of the
            grid boundaries, by default "zero"
        """
        Constituent.__init__(self)
        WavelengthInterpolatorMixin.__init__(
            self,
            wavelengths_nm=wavelengths_nm,
            wavenumbers_cminv=None,
            out_of_bounds_mode=out_of_bounds_mode,
            param_length=len(np.atleast_1d(isotropic)),
        )
        self._iso = np.atleast_1d(isotropic)
        self._vol = np.atleast_1d(volumetric)
        self._geo = np.atleast_1d(geometric)

    @property
    def isotropic(self) -> np.array:
        return self._iso

    @isotropic.setter
    def isotropic(self, iso: np.array):
        self._iso = np.atleast_1d(iso)

    @property
    def volumetric(self) -> np.array:
        return self._vol

    @volumetric.setter
    def volumetric(self, vol: np.array):
        self._vol = np.atleast_1d(vol)

    @property
    def geometric(self) -> np.array:
        return self._geo

    @geometric.setter
    def geometric(self, geo: np.array):
        self._geo = np.atleast_1d(geo)

    def add_to_atmosphere(self, atmo: Atmosphere):
        if atmo.wavelengths_nm is None:
            msg = "Atmosphere must have wavelengths defined before using MODIS"
            raise ValueError(msg)

        atmo.surface.brdf = MODISStokes_1() if atmo.nstokes == 1 else MODISStokes_3()

        interp_matrix = self._interpolating_matrix(atmo)
        atmo.surface.brdf_args[0, :] = interp_matrix @ self._iso
        atmo.surface.brdf_args[1, :] = interp_matrix @ self._vol
        atmo.surface.brdf_args[2, :] = interp_matrix @ self._geo

    def register_derivative(self, atmo: Atmosphere, name: str):
        # TODO update once C++ derivatives are implemented
        # Start by constructing the interpolation matrix
        interp_matrix = self._interpolating_matrix(atmo)

        derivs = {}

        iso_deriv = np.zeros((atmo.num_wavel, 3))
        iso_deriv[:, 0] = 1

        deriv_mapping = atmo.surface.get_derivative_mapping(f"wf_{name}_isotropic")
        deriv_mapping.d_brdf[:] = iso_deriv
        deriv_mapping.interpolator = interp_matrix
        deriv_mapping.interp_dim = f"{name}_wavelength"

        vol_deriv = np.zeros((atmo.num_wavel, 3))
        vol_deriv[:, 1] = 1

        deriv_mapping = atmo.surface.get_derivative_mapping(f"wf_{name}_volumetric")
        deriv_mapping.d_brdf[:] = vol_deriv
        deriv_mapping.interpolator = interp_matrix
        deriv_mapping.interp_dim = f"{name}_wavelength"

        geo_deriv = np.zeros((atmo.num_wavel, 3))
        geo_deriv[:, 2] = 1

        deriv_mapping = atmo.surface.get_derivative_mapping(f"wf_{name}_geometric")
        deriv_mapping.d_brdf[:] = geo_deriv
        deriv_mapping.interpolator = interp_matrix
        deriv_mapping.interp_dim = f"{name}_wavelength"

        return derivs



from __future__ import annotations

from sasktran2.atmosphere import Atmosphere
from sasktran2.solar import SolarModel

from .base import Constituent


class SolarIrradiance(Constituent):
    def __init__(
        self,
        **kwargs,
    ) -> None:
        """
        A constituent that calculates the solar irradiance at TOA using a reference solar irradiance
        spectrum.
        """
        self._model = SolarModel(**kwargs)

    def add_to_atmosphere(self, atmo: Atmosphere):
        if atmo.wavelengths_nm is None:
            msg = "It is required to give the Atmosphere object wavelengths to use the SolarIrradiance constituent"
            raise ValueError(msg)

        atmo.storage.solar_irradiance = self._model.irradiance(atmo.wavelengths_nm)

    def register_derivative(self, atmo, name):
        pass



from __future__ import annotations

import numpy as np


def linear_interpolating_matrix(
    from_grid: np.array, to_grid: np.array, out_of_bounds_mode: str
) -> np.ndarray:
    """
    Constructs an interpolating matrix that can be used to interpolate from one grid to another.

    Parameters
    ----------
    from_grid : np.array
        Grid with the original values. Must be sorted in ascending order
    to_grid : np.array
        Grid we are going to interpolate to. Must be sorted in ascending order
    out_of_bounds_mode : str
        One of ["zero", "extend"], "zero" will set the result to 0 outside the bounds of the original grid,
        "extend" will extend the lowest and highest values

    Returns
    -------
    np.ndarray
        A matrix such that M @ (values on from_grid) = (values on to_grid)
    """
    M = np.zeros((len(to_grid), len(from_grid)))

    for idx, ele in enumerate(to_grid):
        if ele < from_grid[0] or ele > from_grid[-1]:
            if out_of_bounds_mode == "zero":
                continue
            if out_of_bounds_mode == "extend":
                if ele < from_grid[0]:
                    M[idx, 0] = 1
                else:
                    M[idx, -1] = 1
                continue
            msg = f"Invalid out_of_bounds_mode {out_of_bounds_mode}"
            raise ValueError(msg)

        # else we are inside the grid
        idx_above = np.nonzero(from_grid > ele)[0]
        idx_above = len(from_grid) if len(idx_above) == 0 else idx_above[0]

        interp_ele = ele
        if idx_above == 0:
            idx_above = 1
            interp_ele = from_grid[0]

        if idx_above == len(from_grid):
            M[idx, idx_above - 1] = 1
        else:
            w = (from_grid[idx_above] - interp_ele) / (
                from_grid[idx_above] - from_grid[idx_above - 1]
            )
            M[idx, idx_above] = 1 - w
            M[idx, idx_above - 1] = w

    return M



from __future__ import annotations

import logging

import numpy as np

from sasktran2._core import LinearizedMie
from sasktran2.atmosphere import Atmosphere
from sasktran2.mie.distribution import ParticleSizeDistribution, integrate_mie
from sasktran2.mie.refractive import RefractiveIndex
from sasktran2.optical.base import OpticalProperty, OpticalQuantities
from sasktran2.polarization import LegendreStorageView


class Mie(OpticalProperty):
    def __init__(
        self,
        psize_distribution: ParticleSizeDistribution,
        refractive_index: RefractiveIndex,
    ):
        """
        Mie scattering optical property where the Mie calculations are done on the fly.

        Note, this is much slower than using a precomputed Mie table, and in most cases it is
        recommended to use the MieDatabase class instead.

        Parameters
        ----------
        psize_distribution : ParticleSizeDistribution
            Particle size distribution to use for the Mie calculations
        refractive_index : RefractiveIndex
            Refraction index to use for the Mie calculations
        """
        self._psize_distribution = psize_distribution

        self._geometry_dependent = len(psize_distribution.args()) > 0

        self._refrac_index = refractive_index

        self._calculation_ds = {}
        self._xs_ds = {}

    def _internal_generate(self, wavelengths_nm, result_ds, **kwargs):
        if len(wavelengths_nm) > 20:
            logging.warning(
                "Calculating Mie scattering parameters for a large number of wavelengths. You may want to use the MieDatabase class instead."
            )

        dist_args = self._psize_distribution.args()

        compute_coeff = "num_legendre" in kwargs

        for arg in dist_args:
            if arg not in kwargs:
                msg = f"Missing argument {arg} for particle size distribution"
                raise ValueError(msg)

        if self._geometry_dependent:
            # Find the unique values of the distribution arguments
            arg_tuples = np.array([kwargs[arg] for arg in dist_args]).T
            unique_args = np.unique(arg_tuples, axis=0)

            if len(unique_args) > 20:
                logging.warning(
                    "Calculating Mie scattering parameters for a large number of particle size distribution arguments. You may want to use the MieDatabase class instead."
                )

            for args in unique_args:
                # Generate the Mie scattering cross sections for each unique set of distribution arguments
                ds = integrate_mie(
                    LinearizedMie(kwargs.get("num_threads", 1)),
                    self._psize_distribution.distribution(
                        **dict(zip(dist_args, args, strict=True))
                    ),
                    self._refrac_index.refractive_index_fn,
                    wavelengths_nm,
                    compute_coeffs=compute_coeff,
                    num_coeffs=kwargs.get("num_legendre", 10),
                )

                result_ds[tuple(args)] = ds
        else:
            ds = integrate_mie(
                LinearizedMie(kwargs.get("num_threads", 1)),
                self._psize_distribution.distribution(),
                self._refrac_index.refractive_index_fn,
                wavelengths_nm,
                compute_coeffs=compute_coeff,
                num_coeffs=kwargs.get("num_legendre", 10),
            )

            result_ds[()] = ds

    def atmosphere_quantities(self, atmo: Atmosphere, **kwargs) -> OpticalQuantities:
        self._internal_generate(
            atmo.wavelengths_nm,
            self._calculation_ds,
            **{
                **kwargs,
                "num_legendre": atmo.leg_coeff.a1.shape[0],
                "num_threads": atmo._config.num_threads,
            },
        )

        # Then calculate the atmosphere quantities
        quants = OpticalQuantities(
            extinction=np.zeros(
                (len(atmo.model_geometry.altitudes()), len(atmo.wavelengths_nm))
            ),
            ssa=np.zeros(
                (len(atmo.model_geometry.altitudes()), len(atmo.wavelengths_nm))
            ),
        )

        quants.leg_coeff = np.zeros_like(atmo.storage.leg_coeff)

        leg_coeff = LegendreStorageView(quants.leg_coeff, atmo.nstokes)

        if self._geometry_dependent:
            arg_tuples = np.array(
                [kwargs[arg] for arg in self._psize_distribution.args()]
            ).T
            for i, arg in enumerate(arg_tuples):
                ds = self._calculation_ds[tuple(arg)].sel(
                    wavelength=atmo.wavelengths_nm
                )

                # Convert nm^2 to m^2
                quants.extinction[i, :] = ds["xs_total"].to_numpy() / 1e18
                quants.ssa[i, :] = ds["xs_scattering"].to_numpy() / 1e18

                leg_coeff.a1[:, i, :] = ds["lm_a1"].to_numpy().T
                if atmo.nstokes == 3:
                    leg_coeff.a2[:, i, :] = ds["lm_a2"].to_numpy().T
                    leg_coeff.b1[:, i, :] = ds["lm_b1"].to_numpy().T
                    leg_coeff.a3[:, i, :] = ds["lm_a3"].to_numpy().T
        else:
            ds = self._calculation_ds[()].sel(wavelength=atmo.wavelengths_nm)

            # Convert nm^2 to m^2
            quants.extinction[:] = ds["xs_total"].to_numpy()[np.newaxis, :] / 1e18
            quants.ssa[:] = ds["xs_scattering"].to_numpy()[np.newaxis, :] / 1e18

            leg_coeff.a1[:] = ds["lm_a1"].to_numpy().T[:, np.newaxis, :]
            if atmo.nstokes == 3:
                leg_coeff.a2[:] = ds["lm_a2"].to_numpy().T[:, np.newaxis, :]
                leg_coeff.b1[:] = ds["lm_b1"].to_numpy().T[:, np.newaxis, :]
                leg_coeff.a3[:] = ds["lm_a3"].to_numpy().T[:, np.newaxis, :]

        quants.extinction[np.isnan(quants.extinction)] = 0
        quants.ssa[np.isnan(quants.ssa)] = 0

        return quants

    def cross_sections(self, wavelengths_nm, altitudes_m, **kwargs):
        self._internal_generate(wavelengths_nm, self._xs_ds, **kwargs)

        # Then calculate the atmosphere quantities
        quants = OpticalQuantities(
            extinction=np.zeros((len(altitudes_m), len(wavelengths_nm))),
            ssa=np.zeros((len(altitudes_m), len(wavelengths_nm))),
        )

        if self._geometry_dependent:
            arg_tuples = np.array(
                [kwargs[arg] for arg in self._psize_distribution.args()]
            ).T
            for i, arg in enumerate(arg_tuples):
                ds = self._xs_ds[tuple(arg)].sel(wavelength=wavelengths_nm)

                # Convert nm^2 to m^2
                quants.extinction[i, :] = ds["xs_total"].to_numpy() / 1e18
                quants.ssa[i, :] = (
                    ds["xs_scattering"].to_numpy() / 1e18
                ) / quants.extinction[i, :]

            quants.extinction[np.isnan(quants.extinction)] = 0
            quants.ssa[np.isnan(quants.ssa)] = 0
        else:
            ds = self._xs_ds[()].sel(wavelength=wavelengths_nm)

            # Convert nm^2 to m^2
            quants.extinction[:] = ds["xs_total"].to_numpy() / 1e18
            quants.ssa[:] = (ds["xs_scattering"].to_numpy() / 1e18) / quants.extinction

            quants.extinction[np.isnan(quants.extinction)] = 0
            quants.ssa[np.isnan(quants.ssa)] = 0

        return quants



from __future__ import annotations

import hashlib
import json
from itertools import product
from pathlib import Path

import numpy as np
import xarray as xr

from sasktran2.mie import LinearizedMie
from sasktran2.mie.distribution import ParticleSizeDistribution, integrate_mie
from sasktran2.mie.refractive import RefractiveIndex
from sasktran2.optical.database import OpticalDatabaseGenericScatterer

from .base import CachedDatabase


class MieDatabase(CachedDatabase, OpticalDatabaseGenericScatterer):
    def __init__(
        self,
        psize_distribution: ParticleSizeDistribution,
        refractive_index: RefractiveIndex,
        wavelengths_nm: np.array,
        db_root: Path | None = None,
        backend: str = "sasktran2",
        max_legendre_moments: int = 64,
        **kwargs,
    ) -> None:
        """
        A MieDatabase is a database that caches Mie scattering data on the local file system.
        The root directly can optionally be specified, otherwise the default
        folder is used.

        Parameters
        ----------
        psize_distribution : ParticleSizeDistribution
            The particle size distribution
        refractive_index : RefractiveIndex
            The refractive index
        wavelengths_nm : np.array
            The wavelengths to calculate the Mie parameters at
        db_root : Path, optional
            The root directory to store the database, by default None
        backend : str, optional
            The backend to use, by default "sasktran2". Currently supported
            options are ["sasktran_legacy", "sasktran2]. "sasktran_legacy"
            requires the module `sasktran` to be installed.
        kwargs
            Additional arguments to pass to the particle size distribution, these should match the psize_distribution.args() method
        """

        class NumpyEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, np.ndarray):
                    return obj.tolist()
                return json.JSONEncoder.default(self, obj)

        hasher = hashlib.sha1()
        encoded = json.dumps(
            {
                **kwargs,
                "wavelengths": wavelengths_nm,
                "max_legendre_moments": max_legendre_moments,
            },
            sort_keys=True,
            cls=NumpyEncoder,
        ).encode()
        hasher.update(encoded)
        identifier = hasher.hexdigest()

        CachedDatabase.__init__(
            self,
            db_root=db_root,
            rel_path=Path("mie")
            .joinpath(refractive_index.identifier)
            .joinpath(psize_distribution.identifier),
        )

        self._data_file = self._db_root.joinpath(f"{identifier}.nc")

        self._psize_args = psize_distribution.args()
        self._psize_dist = psize_distribution
        self._refractive_index = refractive_index
        self._wavelengths_nm = wavelengths_nm
        self._backend = backend
        self._max_legendre_moments = max_legendre_moments

        for arg in kwargs:
            if arg not in self._psize_args:
                msg = f"Invalid argument {arg} for particle size distribution"
                raise ValueError(msg)

        self._kwargs = kwargs

        OpticalDatabaseGenericScatterer.__init__(self, self.path())

    def generate(self):
        if self._backend == "sasktran_legacy":
            self._generate_sasktran_legacy()
        elif self._backend == "sasktran2":
            self._generate_sasktran2()
        else:
            msg = f"Invalid backend {self._backend}"
            raise ValueError(msg)

    def _generate_sasktran_legacy(self):
        """
        Generates the data file from legacy sasktran
        """
        try:
            from sasktran import MieWiscombe
            from sasktran.legendre import compute_greek_coefficients_legendre
            from sasktran.mie.distribution import integrated_mie
        except ImportError as err:
            msg = "sasktran_legacy is required to generate Mie databases, try pip install sasktran"
            raise ImportError(msg) from err

        def _create_table(distribution, refractive_index_fn, wavelengths):
            mie = MieWiscombe(max_legendre_moment=self._max_legendre_moments)

            vals = integrated_mie(
                mie,
                distribution,
                refractive_index_fn,
                wavelengths,
                num_quad=1000,
                maxintquantile=0.99999,
            )

            # vals xs is in units of nm^2, convert to cm^2 then to m^2
            vals["xs_total"] *= 1e-14 * 1e-4
            vals["xs_scattering"] *= 1e-14 * 1e-4
            vals["xs_absorption"] *= 1e-14 * 1e-4

            lm_a1 = np.zeros_like(vals["lm_p11"].values)
            lm_a2 = np.zeros_like(lm_a1)
            lm_a3 = np.zeros_like(lm_a1)
            lm_a4 = np.zeros_like(lm_a1)
            lm_b1 = np.zeros_like(lm_a1)
            lm_b2 = np.zeros_like(lm_a1)

            for idx in range(len(vals.wavelength.values)):
                selected = vals.isel(wavelength=idx)

                (
                    lm_a1[idx, :],
                    lm_a2[idx, :],
                    lm_a3[idx, :],
                    lm_a4[idx, :],
                    lm_b1[idx, :],
                    lm_b2[idx, :],
                ) = compute_greek_coefficients_legendre(
                    selected["lm_p11"].values,
                    selected["lm_p12"].values,
                    selected["lm_p11"].values,
                    selected["lm_p33"].values,
                    selected["lm_p34"].values,
                    selected["lm_p33"].values,
                )

            vals["lm_a1"] = (["wavelength", "legendre"], lm_a1)
            vals["lm_a2"] = (["wavelength", "legendre"], lm_a2)
            vals["lm_a3"] = (["wavelength", "legendre"], lm_a3)
            vals["lm_a4"] = (["wavelength", "legendre"], lm_a4)
            vals["lm_b1"] = (["wavelength", "legendre"], lm_b1)
            vals["lm_b2"] = (["wavelength", "legendre"], lm_b2)

            ret_ds = xr.Dataset()

            ret_ds["xs_scattering"] = vals["xs_scattering"]
            ret_ds["xs_total"] = vals["xs_total"]
            ret_ds["lm_a1"] = vals["lm_a1"]
            ret_ds["lm_a2"] = vals["lm_a2"]
            ret_ds["lm_a3"] = vals["lm_a3"]
            ret_ds["lm_a4"] = vals["lm_a4"]
            ret_ds["lm_b1"] = vals["lm_b1"]
            ret_ds["lm_b2"] = vals["lm_b2"]

            return ret_ds

        refractive = self._refractive_index.refractive_index_fn

        entries = []
        for vals in product(*self._kwargs.values()):
            psize_args = dict(zip(self._kwargs.keys(), vals, strict=True))

            dist = self._psize_dist.distribution(**psize_args)

            entry = _create_table(dist, refractive, self._wavelengths_nm)

            for key, val in psize_args.items():
                entry[key] = val

            entries.append(entry)

        if len(self._kwargs) > 1:
            # Have to do a multi-index unstack
            ds = (
                xr.concat(entries, dim="args")  # noqa: PD010
                .set_index(args=list(self._kwargs.keys()))
                .unstack("args")
                .rename_dims({"wavelength": "wavelength_nm"})
                .rename_vars({"wavelength": "wavelength_nm"})
            )
        elif len(self._kwargs) == 1:
            ds = (
                xr.concat(entries, dim=next(iter(self._kwargs.keys())))
                .rename_dims({"wavelength": "wavelength_nm"})
                .rename_vars({"wavelength": "wavelength_nm"})
            )
        else:
            # length is 0
            ds = (
                entries[0]
                .rename_dims({"wavelength": "wavelength_nm"})
                .rename_vars({"wavelength": "wavelength_nm"})
            )

        ds.to_netcdf(self._data_file)

    def _generate_sasktran2(self):
        """
        Generates the data file from sasktran2
        """

        def _create_table(distribution, refractive_index_fn, wavelengths):

            mie = LinearizedMie()
            vals = integrate_mie(
                mie,
                distribution,
                refractive_index_fn,
                wavelengths,
                num_quad=1000,
                maxintquantile=0.99999,
                compute_coeffs=True,
                num_coeffs=self._max_legendre_moments,
            )
            # vals xs is in units of nm^2, convert to cm^2 then to m^2
            vals["xs_total"] *= 1e-14 * 1e-4
            vals["xs_scattering"] *= 1e-14 * 1e-4
            vals["xs_absorption"] *= 1e-14 * 1e-4

            ret_ds = xr.Dataset()

            ret_ds["xs_scattering"] = vals["xs_scattering"]
            ret_ds["xs_total"] = vals["xs_total"]
            ret_ds["lm_a1"] = vals["lm_a1"]
            ret_ds["lm_a2"] = vals["lm_a2"]
            ret_ds["lm_a3"] = vals["lm_a3"]
            ret_ds["lm_a4"] = vals["lm_a4"]
            ret_ds["lm_b1"] = vals["lm_b1"]
            ret_ds["lm_b2"] = vals["lm_b2"]

            return ret_ds

        refractive = self._refractive_index.refractive_index_fn

        entries = []
        for vals in product(*self._kwargs.values()):
            psize_args = dict(zip(self._kwargs.keys(), vals, strict=True))

            dist = self._psize_dist.distribution(**psize_args)

            entry = _create_table(dist, refractive, self._wavelengths_nm)

            for key, val in psize_args.items():
                entry[key] = val

            entries.append(entry)

        if len(self._kwargs) > 1:
            # Have to do a multi-index unstack
            ds = (
                xr.concat(entries, dim="args")  # noqa: PD010
                .set_index(args=list(self._kwargs.keys()))
                .unstack("args")
                .rename_dims({"wavelength": "wavelength_nm"})
                .rename_vars({"wavelength": "wavelength_nm"})
            )
        elif len(self._kwargs) == 1:
            ds = (
                xr.concat(entries, dim=next(iter(self._kwargs.keys())))
                .rename_dims({"wavelength": "wavelength_nm"})
                .rename_vars({"wavelength": "wavelength_nm"})
            )
        else:
            # length is 0
            ds = (
                entries[0]
                .rename_dims({"wavelength": "wavelength_nm"})
                .rename_vars({"wavelength": "wavelength_nm"})
            )

        ds.to_netcdf(self._data_file)

    def clear(self):
        if self._data_file.exists():
            self._data_file.unlink()

    def path(self, key: str | None = None, **kwargs) -> Path | None:  # noqa: ARG002
        if not self._data_file.exists():
            self.generate()
        return self._data_file

    def load_ds(self, key: str | None = None, **kwargs) -> xr.Dataset:
        return xr.open_dataset(self.path(key, **kwargs))



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path

import numpy as np
import pandas as pd
from scipy import interpolate

from sasktran2.database.web import StandardDatabase


class RefractiveIndex:
    def __init__(
        self, refractive_index_fn: Callable[[float], complex], identifier: str
    ) -> None:
        """
        A generic implementation of a refractive index function. This class is a light wrapper
        on top of a callable function that returns the complex refractive index for a given
        wavelength, as well as providing a unique identifier for the refractive index.

        Parameters
        ----------
        refractive_index_fn : Callable[[float], complex]
            Function that takes in a wavelength in nm and returns the complex refractive index
        identifier : str
            A unique identifier for the refractive index
        """
        self._fn = refractive_index_fn
        self._identifier = identifier

    def refractive_index(self, wavelength_nm: np.ndarray) -> np.ndarray:
        """
        Returns the complex refractive index for a given wavelength

        Parameters
        ----------
        wavelength_nm : np.ndarray

        Returns
        -------
        np.ndarray
        """
        return self._fn(wavelength_nm)

    @property
    def refractive_index_fn(self):
        """
        Get the function that returns the complex refractive index
        """
        return self._fn

    @property
    def identifier(self):
        """
        Get the unique identifier for this refractive index
        """
        return self._identifier


def _from_osiris_file(path: Path):
    """
    Reads in data from the old OSIRIS refractive index files and returns a callable function
    """
    data = pd.read_csv(path.as_posix(), header=None)

    return interpolate.interp1d(
        data.values[:, 0], data.values[:, 1] - 1j * data.values[:, 2]
    )


class H2SO4(RefractiveIndex):
    def __init__(self, source: str = "osiris") -> None:
        """
        A refractive index for H2SO4. The default source is the old OSIRIS data.
        """
        if source.lower() == "osiris":
            super().__init__(
                _from_osiris_file(
                    StandardDatabase().path("refractive_index/refrac_h2so4_osiris.txt")
                ),
                "H2SO4_osiris",
            )
        else:
            msg = "Only osiris source is supported for H2SO4 refractive index data"
            raise ValueError(msg)


class Dust(RefractiveIndex):
    def __init__(self, source: str = "osiris") -> None:
        """
        A refractive index for Dust. The default source is the old OSIRIS data.
        """
        if source.lower() == "osiris":
            super().__init__(
                _from_osiris_file(
                    StandardDatabase().path("refractive_index/refrac_dust_osiris.txt")
                ),
                "dust_osiris",
            )
        else:
            msg = "Only osiris source is supported for dust refractive index data"
            raise ValueError(msg)


class Ice(RefractiveIndex):
    def __init__(self, source: str = "osiris") -> None:
        """
        A refractive index for ice. The default source is the old OSIRIS data.
        """
        if source.lower() == "osiris":
            super().__init__(
                _from_osiris_file(
                    StandardDatabase().path("refractive_index/refrac_ice_osiris.txt")
                ),
                "ice_osiris",
            )
        else:
            msg = "Only osiris source is supported for ice refractive index data"
            raise ValueError(msg)


class Water(RefractiveIndex):
    def __init__(self, source: str = "osiris") -> None:
        """
        A refractive index for Water. The default source is the old OSIRIS data.
        """
        if source.lower() == "osiris":
            super().__init__(
                _from_osiris_file(
                    StandardDatabase().path("refractive_index/refrac_water_osiris.txt")
                ),
                "water_osiris",
            )
        else:
            msg = "Only osiris source is supported for water refractive index data"
            raise ValueError(msg)



from __future__ import annotations

import abc

import pandas as pd

from .model import SolarModel

__all__ = ["SolarModel"]


class SolarGeometryHandlerBase:
    def __init__(self):
        """
        Interface class for the solar handler.  A solar handler is used to calculate the
        solar angles for a give time and location.
        """

    @abc.abstractmethod
    def target_solar_angles(
        self, latitude: float, longitude: float, altitude: float, time: pd.Timestamp
    ) -> tuple[float, float]:
        """
        Calculates the solar zenith and solar azimuth angles for a given location and time.

        Parameters
        ----------
        latitude : float
            Latitude in [degrees north]
        longitude : float
            Longitude in [degrees east]
        altitude : float
            Altius in [m]
        time : pd.Timestamp
            Time of interest

        Returns
        -------
        tuple[float, float]
            Solar zenith and solar azimuth angles in [degrees]. Solar azimuth is relative to true north, not
            to the observer. And is defined clockwise from north, such that the east is 90 degrees, south is 180 degrees, etc.
            The azimuthal direction points towards the sun.
        """


class SolarGeometryHandlerForced(SolarGeometryHandlerBase):
    def __init__(self, solar_zenith: float, solar_azimuth: float):
        """
        A solar handler where the solar angles are forced to be the same for all locations and times.

        Parameters
        ----------
        solar_zenith : float
            Solar zenith angle in [degrees]
        solar_azimuth : float
            Solar azimuth angle in [degrees]. Here 0 degrees corresponds to pointing true north, 90 degrees to east, etc.
            Note that this is NOT relative to the observer.  Note this is also pointing towards the sun, not away from the sun.
        """
        self.solar_zenith = solar_zenith
        self.solar_azimuth = solar_azimuth

    def target_solar_angles(
        self,
        latitude: float,  # noqa: ARG002
        longitude: float,  # noqa: ARG002
        altitude: float,  # noqa: ARG002
        time: pd.Timestamp,  # noqa: ARG002
    ) -> tuple[float, float]:
        return self.solar_zenith, self.solar_azimuth


class SolarGeometryHandlerAstropy(SolarGeometryHandlerBase):
    """
    Solar handler where the astropy package is used to calculate the solar angles.
    Must have astropy installed in the Python environment in order to use this solar handler.
    """

    def target_solar_angles(
        self, latitude: float, longitude: float, altitude: float, time: pd.Timestamp
    ) -> tuple[float, float]:
        try:
            from astropy.coordinates import AltAz, EarthLocation, get_sun
            from astropy.time import Time
        except ImportError:
            msg = "Astropy is required to use the astropy solar hanlder"
            raise ImportError(msg)  # noqa: B904

        # Location object
        loc = EarthLocation(lat=latitude, lon=longitude, height=altitude)
        sun = get_sun(Time(time))

        altaz = sun.transform_to(AltAz(obstime=time, location=loc))

        return 90.0 - altaz.alt.deg, altaz.az.deg



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
import sasktran2 as sk


def test_forced_handler():
    forced_solar_zenith = 45.0
    forced_solar_azimuth = 180.0

    handler = sk.solar.SolarGeometryHandlerForced(
        forced_solar_zenith, forced_solar_azimuth
    )

    latitude = 0
    longitude = 0

    solar_zenith, solar_azimuth = handler.target_solar_angles(
        latitude, longitude, 0.0, pd.Timestamp.now()
    )

    np.testing.assert_allclose(solar_zenith, forced_solar_zenith)
    np.testing.assert_allclose(solar_azimuth, forced_solar_azimuth)


def test_astropy_handler():
    pytest.importorskip("astropy")

    handler = sk.solar.SolarGeometryHandlerAstropy()

    latitude = 39.74
    longitude = -104.99

    time = pd.Timestamp("2024-11-12 20:00:00")

    solar_zenith, solar_azimuth = handler.target_solar_angles(
        latitude, longitude, 0.0, time
    )

    # Values calculated from NOAA calculator, I assume astropy is more accurate
    np.testing.assert_allclose(solar_zenith, 90.0 - 29.65, atol=1e-1)
    np.testing.assert_allclose(solar_azimuth, 200.82, atol=1e-1)



import numpy as np
import xarray as xr
import sasktran2 as sk
import matplotlib.pyplot as plt


# https://hitran.org/data/CIA/O2-O2_2018b.cia

fname_in = "tools/databases/O2-O2_2018b.cia"
fname_out = R"C:\Users\luf542\AppData\Local\usask-arg\sasktran2\database\cross_sections\o2o2\hitran_cia.nc"

# header line numbers
groups = [
    dict(header_line_number=ln)
    for ln in [
        1,
        4004,
        8007,
        12010,
        16013,
        20016,
        24019,
        28022,
        32025,
        36028,
        40031,
        44034,
        48037,
        52040,
        56043,
        60046,
        64301,
        64808,
        65526,
        68099,
        68793,
        70169,
        71545,
        72919,
        74295,
        87387,
        100366,
        113414,
        126550,
    ]
]

# read header data
with open(fname_in, "r") as f:
    lines = f.readlines()
    for g in groups:
        line = lines[g["header_line_number"] - 1]
        tokens = line.split()
        g["size"] = int(tokens[3])
        g["temperature"] = float(tokens[4])
        g["wavenumber_range"] = (float(tokens[1]), float(tokens[2]))

# read wavenumber and cross section data
for g in groups:
    data = np.loadtxt(
        fname=fname_in, skiprows=g["header_line_number"], max_rows=g["size"]
    )
    g["wavenumber"] = data[:, 0]  # cm5
    g["cross_section"] = data[:, 1]

# construct a wavenumber grid that spans all the given ranges

# 0-14: 0.2 spacing (but with 0.1 spacing for 2 data points?)
# 15: ~0.245 spacing
# 16: 1.0 spacing
# 17: 1.0 spacing
# 18: ~0.482 spacing
# 19: 1.0 spacing
# 20-23: 1.0 spacing
# 24-28: 1.0 spacing

# 0-14 have identical grids
# 20-23 have slightly differing bounds, but this is not a problem because all points are integer wavenumbers (just span the integers from overall min to overall max)
# 24-28 have slightly differing bounds, but ^
# some overlap between 20-23 and 24-28, but ^


# pad regions with an extra data point so we return 0 between bands instead of trying to interpolate between
def pad(v: np.ndarray):
    return np.unique([np.min(v) - 0.01, *v, np.max(v) + 0.01])


wn1 = pad(groups[0]["wavenumber"])
wn2 = pad(groups[15]["wavenumber"])
wn3 = pad(groups[16]["wavenumber"])
wn4 = pad(groups[17]["wavenumber"])
wn5 = pad(groups[18]["wavenumber"])
wn6 = pad(groups[19]["wavenumber"])
wn7_min = np.min([np.min(groups[i]["wavenumber"]) for i in range(21, 29)])
wn7_max = np.max([np.max(groups[i]["wavenumber"]) for i in range(21, 29)])
wn7 = pad(np.arange(wn7_min, wn7_max + 0.1, 1.0))
all_wn = np.concatenate((wn1, wn2, wn3, wn4, wn5, wn6, wn7))

# combine all temperatures into common grid
all_t = np.unique([g["temperature"] for g in groups])

# construct single cross section array
all_xs = np.zeros((len(all_t), len(all_wn)))
regions = [
    list(range(15)),
    [15],
    [16],
    [17],
    [18],
    [19],
    list(range(20, 24)),
    list(range(24, 29)),
]
for region in regions:
    t0 = np.array(
        [groups[k]["temperature"] for k in region]
    )  # temperatures of given data
    for i in range(len(all_t)):
        t = all_t[i]
        if t in t0:  # insert data where all_t matches temperatures of given data
            k = np.searchsorted(t0, t, side="left") + region[0]
            wn = groups[k]["wavenumber"]
            xs = groups[k]["cross_section"]
            j = np.searchsorted(all_wn, wn, side="left")
            all_xs[i, j] = xs
        elif t > np.min(t0) and t < np.max(
            t0
        ):  # interpolate where all_t does not match but is bounded by the temperatures of the given data
            # bound indices
            k2 = np.searchsorted(t0, t, side="right")  # right: subscript 2
            k1 = k2 - 1  # left: subscript 1
            # weights
            w2 = (t - t0[k1]) / (t0[k2] - t0[k1])
            w1 = 1.0 - w2

            k1, k2 = k1 + region[0], k2 + region[0]
            wn1 = groups[k1]["wavenumber"]
            wn2 = groups[k2]["wavenumber"]
            wn = np.union1d(wn1, wn2)

            # pad interpolation data for where grids don't overlap
            xs1 = np.zeros_like(wn)
            j1 = np.searchsorted(wn, wn1, side="left")
            xs1[j1] = groups[k1]["cross_section"]
            xs2 = np.zeros_like(wn)
            j2 = np.searchsorted(wn, wn2, side="left")
            xs2[j2] = groups[k2]["cross_section"]

            # interpolate
            j = np.searchsorted(all_wn, wn, side="left")
            all_xs[i, j] = w1 * xs1 + w2 * xs2
        else:  # extend where all_t does not match and is not bounded by the temperatures of the given data
            k = region[0] + (
                len(t0) - 1 if t > np.max(t0) else 0
            )  # copy min temp data if t < min temp, max temp data if t > max temp
            wn = groups[k]["wavenumber"]
            xs = groups[k]["cross_section"]
            j = np.searchsorted(all_wn, wn, side="left")
            all_xs[i, j] = xs

all_xs_wl = all_xs[:, ::-1]
all_wl = 1e7 / all_wn[::-1]
ds = xr.Dataset(
    {"xs": (["temperature", "wavelength_nm"], all_xs_wl / 1e10)},  # cm^5 -> m^5
    coords={
        "temperature": all_t,
        "wavelength_nm": all_wl,  # assuming vacuum wavelength (see note below)
        # "wavelength_nm": sk.optical.air_wavelength_to_vacuum_wavelength(all_wl),
    },
)

ds.to_netcdf(fname_out)

# note on vacuum vs air wavelength:
# the wavelengths for the 477 nm absorption band are taken from Thalman and Volkamer (2013)
# they calibrated their detector using "Solar Flux Atlas from 296 to 1300 nm" (Kurucz, 1984)
# if the atlas is in vacuum wavelengths, then the o2o2 cross sections are in vacuum wavelength
# I didn't find a copy of the atlas, but I am assuming it is in vacuum wavelength



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def test_air_vacuum_conversions():
    """
    Verifies that if we convert to vacuum and back, we get the same thing
    """
    wavel = np.arange(280, 10000, 0.1)

    conv_back = sk.optical.air_wavelength_to_vacuum_wavelength(
        sk.optical.vacuum_wavelength_to_air_wavelength(wavel)
    )

    np.testing.assert_allclose(wavel, conv_back)



from __future__ import annotations

import logging
from copy import copy
from dataclasses import dataclass

import numpy as np
import xarray as xr
from scipy import sparse

import sasktran2 as sk
from sasktran2.polarization import LegendreStorageView
from sasktran2.units import (
    wavenumber_cminv_to_wavlength_nm,
    wavlength_nm_to_wavenumber_cminv,
)
from sasktran2.util.state import EquationOfState


@dataclass
class NativeGridDerivative:
    """
    Internal input object that defines the model input quantities necessary to calculate a derivative.
    This mapping is from the native model grid to the native grid, it does not change the gridding.
    """

    d_extinction: np.ndarray = None
    d_ssa: np.ndarray = None
    d_leg_coeff: np.ndarray = None
    scat_factor: np.ndarray = None
    scat_deriv_index: int = None
    d_brdf: np.ndarray = None


class DerivativeMapping:
    def __init__(
        self,
        native_grid_mapping: NativeGridDerivative,
        summable: bool = False,
        log_radiance_space: bool = False,
        name_prefix: str = "wf_",
    ):
        """
        A class which defines a mapping from the internal model derivative quantities (derivatives with respect to,
        extinction, single scatter albedo, legendre coefficients) and user input quantities (e.g. VMR at a specific level).

        Parameters
        ----------
        native_grid_mapping : NativeGridDerivative
            A mapping from the native atmospheric grid to the same grid.
        summable : bool, optional
            True if this quantity should be accumulated and summed across each constituent. For example, each constituent may have a derivative with respect
            to a quantity like atmospheric temperature or pressure, by default False
        log_radiance_space : bool, optional
            True if the derivative should be returned in log_radiance space, i.e. if we should divide the absolute derivative by the radiance
        name_prefix: str, optional
            Prefix to put infront of the variable in the output dataset, default is "wf_".
        """
        self._native_grid_mapping = native_grid_mapping
        self._summable = summable
        self._log_radiance_space = log_radiance_space
        self._name_prefix = name_prefix

    @property
    def native_grid_mapping(self) -> NativeGridDerivative:
        """
        The mapping on the native grid.

        Returns
        -------
        NativeGridDerivative
        """
        return self._native_grid_mapping

    @property
    def is_surface_derivative(self) -> bool:
        return self._native_grid_mapping.d_brdf is not None

    @property
    def summable(self) -> bool:
        """
        True if the mapping should be summed and accumulated.

        Returns
        -------
        bool
        """
        return self._summable

    @property
    def log_radiance_space(self) -> bool:
        """
        True if the derivative should be returned in log_radiance space, i.e. if we should divide the absolute derivative by the radiance

        Returns
        -------
        bool
        """
        return self._log_radiance_space

    @property
    def name_prefix(self) -> str:
        return self._name_prefix

    def map_derivative(self, data: np.ndarray, dimensions: list[str]):
        if self.is_surface_derivative:
            return xr.DataArray(
                data.sum(axis=-1),
                dims=dimensions,
            )
        return xr.DataArray(
            data,
            dims=dimensions,
        )

    def name(self) -> str | None:
        return None


class InterpolatedDerivativeMapping(DerivativeMapping):
    def __init__(
        self,
        native_grid_mapping: NativeGridDerivative,
        interpolating_matrix: np.ndarray,
        interp_dim="altitude",
        result_dim="interp_altitude",
        summable: bool = False,
        log_radiance_space: bool = False,
        name: str | None = None,
    ):
        """
        A class which defines a mapping from internal model derivative quantities to user input quantities
        that are not on the native model grid

        Parameters
        ----------
        native_grid_mapping : NativeGridDerivative
            Mapping of the quantity in question from the native grid to the native grid
        interpolating_matrix : np.ndarray
            An interpolating matrix such that user quantity on the native grid can be calculated by multiplying
            the matrix by the user input quantity
        interp_dim : str, optional
            Dimension in the native mapping the interpolation is done over, by default "altitude"
        result_dim : str, optional
            string to name the resulting dimension, by default "interp_altitude"
        summable : bool, optional
            See :py:class:`DerivativeMapping`, by default False
        log_radiance_space : bool, optional
            See :py:class:`DerivativeMapping`, by default False
        name : Optional[str], optional
            A name for the derivative mapping, if set to None then the name is inferred from the key
            the mapping resides in.  This is only useful if you have two derivatives with the same key
            that should be summed together, by default None
        """
        super().__init__(native_grid_mapping, summable, log_radiance_space)
        self._name = name

        if interp_dim != result_dim:
            self._xr_interpolator = xr.DataArray(
                interpolating_matrix, dims=[interp_dim, result_dim]
            )
            self._rename_map = {}
        else:
            self._xr_interpolator = xr.DataArray(
                interpolating_matrix, dims=["tempDIM", result_dim]
            )
            self._rename_map = {interp_dim: "tempDIM"}

        self._sparse_mapping = sparse.csc_matrix(self._xr_interpolator.to_numpy())

    def map_derivative(self, data: np.ndarray, dimensions: list[str]):
        new_dims = copy(dimensions)
        new_dims[-1] = self._xr_interpolator.dims[-1]

        return xr.DataArray(
            (data.reshape(-1, data.shape[-1]) @ self._sparse_mapping).reshape(
                np.concatenate((data.shape[:-1], [-1]))
            ),
            dims=new_dims,
        ).transpose(*[new_dims[-1], *new_dims[:-1]])

    def name(self) -> str | None:
        return self._name


class SurfaceDerivativeMapping(DerivativeMapping):
    def __init__(
        self,
        native_grid_mapping: NativeGridDerivative,
        interpolating_matrix: np.ndarray,
        interp_dim="wavelength",
        result_dim="interp_wavelength",
        summable: bool = False,
        log_radiance_space: bool = False,
    ):
        """
        A class which defines a mapping from internal model surface derivative quantities to user input quantities
        that are not on the native model grid

        Parameters
        ----------
        native_grid_mapping : NativeGridDerivative
            Mapping of the quantity in question from the native grid to the native grid
        interpolating_matrix : np.ndarray
            An interpolating matrix such that user quantity on the native grid can be calculated by multiplying
            the matrix by the user input quantity
        interp_dim : str, optional
            Dimension in the native mapping the interpolation is done over, by default "wavelength"
        result_dim : str, optional
            string to name the resulting dimension, by default "interp_wavelength"
        summable : bool, optional
            See :py:class:`DerivativeMapping`, by default False
        log_radiance_space : bool, optional
            See :py:class:`DerivativeMapping`, by default False
        """
        super().__init__(native_grid_mapping, summable, log_radiance_space)

        self._interp_dim = interp_dim
        self._result_dim = result_dim

        self._xr_interpolator = xr.DataArray(
            interpolating_matrix, dims=["tempDIM", result_dim]
        )

    def map_derivative(self, data: np.ndarray, dimensions: list[str]):
        return xr.DataArray(
            np.einsum(
                "ijk, il->lijk",
                data.sum(axis=-1),
                self._xr_interpolator.to_numpy(),
                optimize=True,
            ),
            dims=[self._result_dim, *dimensions],
        )


class Atmosphere:
    def __init__(
        self,
        model_geometry: sk.Geometry1D,
        config: sk.Config,
        wavelengths_nm: np.array = None,
        wavenumber_cminv: np.array = None,
        numwavel: int | None = None,
        calculate_derivatives: bool = True,
        pressure_derivative: bool = True,
        temperature_derivative: bool = True,
        specific_humidity_derivative: bool = True,
        legendre_derivative: bool = True,
    ):
        """
        The main specification for the atmospheric state.

        See :py:attr:`~storage` for details on how the atmospheric state parameters are stored.

        See :py:attr:`~surface` for details on how the surface parameters are stored.

        Parameters
        ----------
        model_geometry : sk.Geometry1D
            The geometry defining where the atmospheric quantities are specified on.
        config : sk.Config
            Main configuration object.
        wavelengths_nm : np.array, optional
            Array of wavelengths to specify the atmosphere at in [nm].  One of wavelengths_nm, wavenumber_cminv, numwavel must be set.
        wavenumber_cminv : np.array, optional
            Array of wavenumbers to specify the atmosphere at in [:math:`\\text{cm}^{-1}`]. One of wavelengths_nm, wavenumber_cminv, numwavel must be set.
        numwavel : int, optional
            Number of wavelengths to include in the calculation.  Note that wavelength is a dummy variable, this dimension can be
            anaything. One of wavelengths_nm, wavenumber_cminv, numwavel must be set.
        calculate_derivatives : bool, optional
            Whether or not the model should calculate derivatives with respect to atmospheric quantities., by default True
        pressure_derivative: bool, optional
            Whether or not the model should calculate derivatives with respect to pressure., by default True
        temperature_derivative: bool, optional
            Whether or not the model should calculate derivatives with respect to temperature., by default True
        legendre_derivative: bool, optional
            Whether or not the model should calculate derivatives with respect to the legendre coefficients., by default True
        """
        self._wavelengths_nm = None
        self._wavenumbers_cminv = None

        # Set wavelengths
        if (
            (wavelengths_nm is None)
            and (wavenumber_cminv is None)
            and (numwavel is None)
        ):
            msg = "One of wavelengths_nm, wavenumber_cminv, numwavel must be set when constructing the Atmosphere object"
            raise ValueError(msg)

        if wavelengths_nm is not None:
            self.wavelengths_nm = wavelengths_nm

        if wavenumber_cminv is not None:
            self.wavenumbers_cminv = wavenumber_cminv

        nwavel = len(self.wavelengths_nm) if numwavel is None else numwavel

        self._nstokes = config.num_stokes
        self._calculate_derivatives = calculate_derivatives
        self._pressure_derivative = pressure_derivative
        self._temperature_derivative = temperature_derivative
        self._legendre_derivative = legendre_derivative
        self._specific_humidity_derivative = specific_humidity_derivative

        if self._nstokes == 1:
            self._atmosphere = sk.AtmosphereStokes_1(
                nwavel, model_geometry, config, calculate_derivatives
            )
        elif self._nstokes == 3:
            self._atmosphere = sk.AtmosphereStokes_3(
                nwavel, model_geometry, config, calculate_derivatives
            )

        self._leg_coeff = LegendreStorageView(
            self._atmosphere.storage.leg_coeff, self._nstokes
        )

        self._storage_needs_reset = False

        self._equation_of_state = EquationOfState()

        self._model_geometry = model_geometry
        self._config = config

        self._constituents = {}
        self._derivs = {}
        self._unscaled_ssa = None
        self._unscaled_extinction = None
        self._applied_delta_m_order = None
        self._nwavel = nwavel

    @property
    def num_wavel(self) -> int:
        """
        The number of wavelengths the atmosphere is specified at

        Returns
        -------
        int
        """
        return self._nwavel

    @property
    def model_geometry(self) -> sk.Geometry1D:
        """
        The model geometry object

        Returns
        -------
        sk.Geometry1D
        """
        return self._model_geometry

    @property
    def applied_delta_m_order(self) -> int | None:
        """
        The order of the applied delta_m scaling. Can be None if no scaling has been applied
        """
        return self._applied_delta_m_order

    @property
    def nstokes(self) -> int:
        """
        The number of stokes parameters the atmosphere contains information to calculate

        Returns
        -------
        int
        """
        return self._nstokes

    @property
    def calculate_temperature_derivative(self) -> bool:
        """
        True if we are calculating the derivative with respect to temperature

        Returns
        -------
        bool
        """
        return self._temperature_derivative

    @property
    def calculate_pressure_derivative(self) -> bool:
        """
        True if we are calculating the derivative with respect to pressure

        Returns
        -------
        bool
        """
        return self._pressure_derivative

    @property
    def calculate_specific_humidity_derivative(self) -> bool:
        """
        True if we are calculating the derivative with respect to specific humidity

        Returns
        -------
        bool
        """
        return self._specific_humidity_derivative

    @property
    def storage(
        self,
    ) -> sk.AtmosphereStorageStokes_1 | sk.AtmosphereStorageStokes_3:
        """
        The internal object which contains the atmosphere extinction, single scatter albedo, and legendre coefficients.

        Returns
        -------
        Union[sk.AtmosphereGridStorageStokes_1, sk.AtmosphereGridStorageStokes_3]
        """
        return self._atmosphere.storage

    @property
    def surface(self) -> sk.SurfaceStokes_1 | sk.SurfaceStokes_3:
        """
        The surface object

        Returns
        -------
        sk.Surface
        """
        return self._atmosphere.surface

    @property
    def temperature_k(self) -> np.array:
        """
        Atmospheric temperature in [K] on the same grid as :py:attr:`~model_geometry`

        Returns
        -------
        np.array
        """
        return self._equation_of_state.temperature_k

    @temperature_k.setter
    def temperature_k(self, temp: np.array):
        self._equation_of_state.temperature_k = temp

    @property
    def pressure_pa(self) -> np.array:
        """
        Pressure in [Pa] on the same grid as :py:attr:`~model_geometry`

        Returns
        -------
        np.array
        """
        return self._equation_of_state.pressure_pa

    @property
    def specific_humidity(self) -> np.array:
        """
        Specific humidity on the same grid as :py:attr:`~model_geometry`

        Returns
        -------
        np.array
        """
        return self._equation_of_state.specific_humidity

    @property
    def state_equation(self) -> EquationOfState:
        """
        The equation of state object which contains the temperature, pressure, and specific humidity

        Returns
        -------
        EquationOfState
        """
        return self._equation_of_state

    @specific_humidity.setter
    def specific_humidity(self, sh: np.array):
        self._equation_of_state.specific_humidity = sh

    @pressure_pa.setter
    def pressure_pa(self, pres: np.array):
        self._equation_of_state.pressure_pa = pres

    @property
    def wavelengths_nm(self) -> np.ndarray | None:
        """
        The wavelengths in [nm] the atmosphere is specified at.  This is an
        optional property, it may be None.

        Returns
        -------
        Optional[np.array]
        """
        return self._wavelengths_nm

    @wavelengths_nm.setter
    def wavelengths_nm(self, wav: np.array):
        self._wavelengths_nm = wav
        self._wavenumbers_cminv = wavlength_nm_to_wavenumber_cminv(wav)

    @property
    def wavenumbers_cminv(self) -> np.ndarray | None:
        """
        The wavenumbers in [:math:`\\text{cm}^{-1}`].  This is an optional property, it may be set to None

        Returns
        -------
        Optional[np.array]
        """
        return self._wavenumbers_cminv

    @wavenumbers_cminv.setter
    def wavenumbers_cminv(self, wav: np.array):
        self._wavenumbers_cminv = wav
        self._wavelengths_nm = wavenumber_cminv_to_wavlength_nm(wav)

    def _zero_storage(self):
        """
        Sets the internal storage object to 0.  Typically used in-between calculations to reset the
        atmosphere without reconstructing it.
        """
        self.storage.ssa[:] = 0
        self.storage.total_extinction[:] = 0
        self.storage.leg_coeff[:] = 0
        self.storage.emission_source[:] = 0

        for _, ele in self.storage.derivative_mappings.items():
            ele.set_zero()
        for _, ele in self.surface.derivative_mappings.items():
            ele.set_zero()

        self.surface.emission[:] = 0

    @property
    def deriv_mappings(self) -> dict:
        """
        A nested dictionary of :py:class:`sasktran2.atmosphere.DerivativeMapping` objects.

        Returns
        -------
        dict
        """
        return self._derivs

    @property
    def unscaled_ssa(self) -> np.ndarray:
        """
        The unscaled single scatter albedo.  After performing the calculation, `storage.ssa` may be
        modified because of the delta-m scaling.  This property stores the original unscaled single
        scatter albedo.

        Returns
        -------
        np.ndarray
        """
        return self._unscaled_ssa

    @property
    def unscaled_extinction(self) -> np.ndarray:
        """
        The unscaled extinction. After performing the calculation, `storage.total_extinction` may be
        modified because of the delta-m scaling. This property stores the original unscaled single
        scatter albedo.

        Returns
        -------
        np.ndarray
        """
        return self._unscaled_extinction

    @property
    def leg_coeff(self) -> LegendreStorageView:
        return self._leg_coeff

    def internal_object(self) -> sk.AtmosphereStokes_1 | sk.AtmosphereStokes_3:
        """
        The internal `pybind11` object that can be used to perform the radiative transfer calculation.
        Calling this method will trigger a construction of the atmosphere object if necessary, and then
        return back the internal object.

        Returns
        -------
        Union[sk.AtmosphereStokes_1, sk.AtmosphereStokes_3]
        """
        num_scat_derivs = 0

        if len(self._constituents) > 0:
            logging.debug("Setting atmosphere from constituents")
            # Using the constituent interface
            if self._storage_needs_reset:
                self._zero_storage()
            for _, constituent in self._constituents.items():
                constituent.add_to_atmosphere(self)

            nz = self.storage.ssa > 0

            self.storage.leg_coeff[:, nz] /= self.storage.ssa[np.newaxis, nz]
            self.storage.ssa[nz] /= self.storage.total_extinction[nz]

            self._derivs = {}
            if self._calculate_derivatives:
                for name, constituent in self._constituents.items():
                    constituent.register_derivative(self, name)

                for _, deriv in self.storage.derivative_mappings.items():
                    if deriv.is_scattering_derivative:
                        num_scat_derivs += 1
            logging.debug("Finished setting atmosphere from constituents")
        else:
            # using the raw interface
            if self._calculate_derivatives and len(self._derivs) == 0:
                deriv_mapping = self.storage.get_derivative_mapping("wf_extinction")
                deriv_mapping.d_extinction[:] = 1
                deriv_mapping.d_ssa[:] = 0.0

                deriv_mapping = self.storage.get_derivative_mapping("wf_ssa")
                deriv_mapping.d_extinction[:] = 0.0
                deriv_mapping.d_ssa[:] = 1.0

                deriv_mapping = self.surface.get_derivative_mapping("wf_albedo")
                deriv_mapping.d_brdf[:] = 1.0

                if self._legendre_derivative:
                    for i in range(self.storage.leg_coeff.shape[0]):
                        if i == 0:
                            # No derivative for the first leg_coeff
                            continue

                        deriv_mapping = self.storage.get_derivative_mapping(
                            f"wf_leg_coeff_{i}"
                        )
                        deriv_mapping.d_leg_coeff[i, :] = 1
                        deriv_mapping.d_extinction[:] = 0
                        deriv_mapping.d_ssa[:] = 0
                        deriv_mapping.scat_factor[:] = 1
                        num_scat_derivs += 1

        # Now we need to resize the phase derivative storage if necessary, and set the scattering derivatives
        if num_scat_derivs > 0:
            self.storage.resize_derivatives(num_scat_derivs)

            scat_index = 0
            for _, mapping in self.storage.derivative_mappings.items():
                if mapping.is_scattering_derivative:
                    self.storage.d_leg_coeff[:, :, :, scat_index] = mapping.d_leg_coeff
                    mapping.scat_deriv_index = scat_index
                    scat_index += 1

        # Store the unscaled optical properties for use in the derivative mappings
        self._unscaled_ssa = copy(self.storage.ssa)
        self._unscaled_extinction = copy(self.storage.total_extinction)

        # Apply delta scaling if required
        if self._config.delta_m_scaling:
            # Find the order of the scaling
            if self._config.num_streams == self._leg_coeff.a1.shape[0]:
                # Can't apply scaling
                logging.info(
                    "Delta-m Scaling NOT applied since the number of MS streams is equal to the number of user input legendre coefficients"
                )
            else:
                self._atmosphere.apply_delta_m_scaling(self._config.num_streams)
                self._applied_delta_m_order = self._config.num_streams

        self._storage_needs_reset = True
        return self._atmosphere

    def __setitem__(self, item, value):
        self._constituents[item] = value

    def __getitem__(self, item):
        return self._constituents.get(item)



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def test_mie_simple_database():
    _ = sk.database.MieDatabase(
        sk.mie.distribution.LogNormalDistribution(),
        sk.mie.refractive.H2SO4(),
        np.array([532, 1020]),
        median_radius=np.array([100, 200]),
        mode_width=np.array([1.5, 1.6, 1.7]),
    )


def test_mie_db_in_engine():
    config = sk.Config()

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    wavel = np.arange(280.0, 800.0, 0.1)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()
    atmosphere["ozone"] = sk.climatology.mipas.constituent("O3", sk.optical.O3DBM())
    atmosphere["no2"] = sk.climatology.mipas.constituent(
        "NO2", sk.optical.NO2Vandaele()
    )

    mie_db = sk.database.MieDatabase(
        sk.mie.distribution.LogNormalDistribution().freeze(
            median_radius=80, mode_width=1.6
        ),
        sk.mie.refractive.H2SO4(),
        wavelengths_nm=np.arange(270, 1000, 50.0),
    )

    aero_ext = np.zeros(len(model_geometry.altitudes()))
    aero_ext[0:30] = 1e-7
    atmosphere["aerosol"] = sk.constituent.ExtinctionScatterer(
        mie_db,
        altitudes_m=model_geometry.altitudes(),
        extinction_per_m=aero_ext,
        extinction_wavelength_nm=745,
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)
    _ = engine.calculate_radiance(atmosphere)


def test_mie_database_against_online():
    config = sk.Config()
    config.num_stokes = 3
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    wavel = np.array([750])
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()
    atmosphere["ozone"] = sk.climatology.mipas.constituent("O3", sk.optical.O3DBM())
    atmosphere["no2"] = sk.climatology.mipas.constituent(
        "NO2", sk.optical.NO2Vandaele()
    )

    mie_db = sk.database.MieDatabase(
        sk.mie.distribution.LogNormalDistribution().freeze(
            median_radius=80, mode_width=1.6
        ),
        sk.mie.refractive.H2SO4(),
        wavelengths_nm=np.array([700, 750, 800]),
    )

    aero_ext = np.zeros(len(model_geometry.altitudes()))
    aero_ext[0:30] = 1e-7
    atmosphere["aerosol"] = sk.constituent.ExtinctionScatterer(
        mie_db,
        altitudes_m=model_geometry.altitudes(),
        extinction_per_m=aero_ext,
        extinction_wavelength_nm=750,
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)
    rad_with_db = engine.calculate_radiance(atmosphere)

    atmosphere["aerosol"] = sk.constituent.ExtinctionScatterer(
        sk.optical.Mie(
            sk.mie.distribution.LogNormalDistribution().freeze(
                median_radius=80, mode_width=1.6
            ),
            sk.mie.refractive.H2SO4(),
        ),
        altitudes_m=model_geometry.altitudes(),
        extinction_per_m=aero_ext,
        extinction_wavelength_nm=750,
    )

    rad_without_db = engine.calculate_radiance(atmosphere)

    p_diff = (
        np.abs(
            rad_with_db["radiance"].to_numpy() - rad_without_db["radiance"].to_numpy()
        )
        / rad_with_db["radiance"].to_numpy()
        * 100
    )

    np.testing.assert_array_almost_equal(p_diff, 0.0, decimal=3)



from __future__ import annotations

import numpy as np

from sasktran2.optical import pressure_temperature_to_numberdensity
from sasktran2.units import celsius_to_kelvin


def _o2_refrac_bates(wavelengths_um: np.array) -> np.array:
    """
    Calculates the O2 refractive index using the parameterization suggested by Bates.

    Parameters
    ----------
    wavelengths_nm : np.array
       Wavelengths in [nm]

    Returns
    -------
    np.array
       O2 Refractive index, (n-1) * 10**8
    """
    coeff_ranges = [
        {"coeff": [23796.7, 168988.4], "range": [0, 0.221]},
        {"coeff": [22120.4, 203187.6], "range": [0.221, 0.288]},
        {"coeff": [20564.8, 248089.9], "range": [0.288, 0.546]},
        {"coeff": [21351.1, 218567.0], "range": [0.546, np.inf]},
    ]

    result = np.zeros_like(wavelengths_um)

    for ele in coeff_ranges:
        good = (wavelengths_um > ele["range"][0]) & (wavelengths_um <= ele["range"][1])

        result[good] = ele["coeff"][0] + ele["coeff"][1] / (
            40.9 - wavelengths_um[good] ** (-2)
        )

    return result


def _n2_refrac_bates(wavelengths_um: np.array) -> np.array:
    """
    Calculates the N2 refractive index using the parameterization suggested by Bates.

    Parameters
    ----------
    wavelengths_nm : np.array
       Wavelengths in [nm]

    Returns
    -------
    np.array
       N2 Refractive index, (n-1) * 10**8
    """
    coeff_ranges = [
        {"coeff": [6998.749, 3233582.0], "range": [0, 0.254]},
        {"coeff": [5989.242, 3363266.3], "range": [0.254, 0.468]},
        {"coeff": [6855.200, 3243157.0], "range": [0.468, np.inf]},
    ]

    result = np.zeros_like(wavelengths_um)

    for ele in coeff_ranges:
        good = (wavelengths_um > ele["range"][0]) & (wavelengths_um <= ele["range"][1])
        delta_lambda = 0.468 - wavelengths_um[good]

        result[good] = (
            ele["coeff"][0]
            + ele["coeff"][1] / (144 - wavelengths_um[good] ** (-2))
            + 2.27684009 * np.sign(delta_lambda) * np.exp(-np.abs(delta_lambda) / 0.003)
        )

    return result


def _ar_refrac_bates(wavelengths_um: np.array) -> np.array:
    """
    Calculates the Ar refractive index using the parameterization suggested by Bates.

    Parameters
    ----------
    wavelengths_nm : np.array
       Wavelengths in [nm]

    Returns
    -------
    np.array
       Ar Refractive index, (n-1) * 10**8
    """
    nsq_m_1 = 5.547e-4 * (
        1 + 5.15e-3 * wavelengths_um ** (-2) + 4.19e-5 * wavelengths_um ** (-4)
    )

    return (np.sqrt(nsq_m_1 + 1) - 1) * 1e8


def _co2_refrac_bates(wavelengths_um: np.array) -> np.array:
    """
    Calculates the CO2 refractive index using the parameterization suggested by Bates.

    Parameters
    ----------
    wavelengths_nm : np.array
       Wavelengths in [nm]

    Returns
    -------
    np.array
       CO2 Refractive index, (n-1) * 10**8
    """
    return (
        22822.1
        + 117.8 * wavelengths_um ** (-2)
        + 2406030 / (130 - wavelengths_um ** (-2))
        + 15997 / (38.9 - wavelengths_um ** (-2))
    )


def _o2_king_bates(wavelengths_um: np.array) -> np.array:
    """
    Calculates the O2 King factor using the parameterization suggested by Bates.

    Parameters
    ----------
    wavelengths_nm : np.array
       Wavelengths in [nm]

    Returns
    -------
    np.array
       O2 King factor
    """
    return 1.096 + 1.385e-3 * wavelengths_um ** (-2) + 1.448e-4 * wavelengths_um ** (-4)


def _n2_king_bates(wavelengths_um: np.array) -> np.array:
    """
    Calculates the N2 King factor using the parameterization suggested by Bates.

    Parameters
    ----------
    wavelengths_nm : np.array
       Wavelengths in [nm]

    Returns
    -------
    np.array
       N2 King factor
    """
    return 1.034 + 3.17e-4 * wavelengths_um ** (-2)


def _ar_king_bates(wavelengths_um: np.array) -> np.array:
    """
    Calculates the Ar King factor using the parameterization suggested by Bates.

    Parameters
    ----------
    wavelengths_nm : np.array
       Wavelengths in [nm]

    Returns
    -------
    np.array
       Ar King factor
    """
    return np.ones_like(wavelengths_um)


def _co2_king_bates(wavelengths_um: np.array) -> np.array:
    """
    Calculates the CO2 King factor using the parameterization suggested by Bates.

    Parameters
    ----------
    wavelengths_nm : np.array
       Wavelengths in [nm]

    Returns
    -------
    np.array
       CO2 King factor
    """
    return np.ones_like(wavelengths_um) * 1.15


def rayleigh_cross_section_bates(
    wavelengths_um: np.array,
    n2_percentage: float = 78.084,
    o2_percentage: float = 20.946,
    ar_percentage: float = 0.934,
    co2_percentage: float = 0.036,
) -> np.array:
    """
    Calculates the Rayleigh scattering cross section in units of [m**2] for a given refractive index according to the Bates
    approximation.  Cross section is returned back at (0C, 1013.25 hPa)


    Parameters
    ----------
    wavelengths_um : np.array
        Wavelengths in [um]
    n2_percentage : float, Optional
        Percentage of N2, default 78.084%
    o2_percentage : float, Optional
        Percentage of O2, default 20.946%
    ar_percentage : float, Optional
        Percentage of Ar, default 0.934%
    co2_percentage : float, Optional
        Percentaage of CO2, default 0.036%

    Returns
    -------
    np.array
        The rayleigh scattering cross section in units [m**2]
    np.array
        The effective King factor [unitless]
    """
    lorenz_factors = (
        o2_percentage
        / 100
        * _o2_refrac_bates(wavelengths_um) ** 2
        * _o2_king_bates(wavelengths_um)
        + n2_percentage
        / 100
        * _n2_refrac_bates(wavelengths_um) ** 2
        * _n2_king_bates(wavelengths_um)
        + ar_percentage
        / 100
        * _ar_refrac_bates(wavelengths_um) ** 2
        * _ar_king_bates(wavelengths_um)
        + co2_percentage
        / 100
        * _co2_refrac_bates(wavelengths_um) ** 2
        * _co2_king_bates(wavelengths_um)
    )

    eff_king = (
        o2_percentage / 100 * _o2_king_bates(wavelengths_um)
        + n2_percentage / 100 * _n2_king_bates(wavelengths_um)
        + ar_percentage / 100 * _ar_king_bates(wavelengths_um)
        + co2_percentage / 100 * _co2_king_bates(wavelengths_um)
    )

    num_dens = pressure_temperature_to_numberdensity(
        1013.25 * 100, celsius_to_kelvin(0)
    )

    return (
        32 * np.pi**3 / (3 * num_dens**2 * wavelengths_um**4) * (lorenz_factors) * 1e8,
        eff_king,
    )



from __future__ import annotations

import abc
import logging
import time

import numpy as np
import scipy.integrate as integrate
import xarray as xr
from scipy.stats import lognorm, rv_continuous

from sasktran2 import mie
from sasktran2.legendre import compute_greek_coefficients
from sasktran2.mie import MieOutput


def _post_process(total: dict, wavelengths):
    """
    Internal method to post-process the Mie solution.  This involves transforming the quantities that we integrate over
    to the relevant bulk properties and applying the necessary normalizations.
        Parameters
    ----------
    total : dict
        Dictionary containing calculated and integrated mie parameters
    wavelengths : np.array
        Wavelengths of interest

    """
    k = 2 * np.pi / wavelengths
    c = 4 * np.pi / (2 * k**2 * total["xs_scattering"])

    total["p11"] *= c
    total["p12"] *= c
    total["p33"] *= c
    total["p34"] *= c

    return


def _integrable_parameters(output: MieOutput, pdf, x):
    """
    Helper function to set up the parameters to be integrated

    Parameters
    ----------
    output : MieOutput
        MieOutput structure containing angles, siza parameters, refractive index, and calculated mie parameters
    pdf : np.array
        Probability distribution evaluated for radii x
    x : np.array
        Radii of the distribution
    Returns
    -------
    Dict
        Dictionary containing the keys 'p11', 'p12', 'p33', 'p34', 'Cext', 'Csca', and 'Cabs'
    """

    out = {}
    s1 = output.values.S1
    s2 = output.values.S2

    # Phase function elements
    # Note that P22 = P11 and P44 = p33 so we have 6 independent elements
    out["p11"] = np.abs(s1) ** 2 + np.abs(s2) ** 2
    out["p12"] = np.abs(s1) ** 2 - np.abs(s2) ** 2
    out["p33"] = np.real(s1 * np.conj(s2) + s2 * np.conj(s1))
    out["p34"] = np.real(-1j * (s1 * np.conj(s2) - s2 * np.conj(s1)))

    # Cross sections in units of input radius**2
    out["Cext"] = output.values.Qext * np.pi * np.square(x)
    out["Csca"] = output.values.Qsca * np.pi * np.square(x)
    out["Cabs"] = (output.values.Qext - output.values.Qsca) * np.pi * np.square(x)

    for key in ["Cext", "Csca", "Cabs"]:
        out[key] *= pdf

    for key in ["p11", "p12", "p33", "p34"]:
        out[key] *= np.transpose([pdf])

    return out


def integrate_mie(
    my_mie: mie,
    prob_dist: rv_continuous,
    refrac_index_fn,
    wavelengths,
    num_angles=1801,
    num_quad=1024,
    maxintquantile=0.99999,
    compute_coeffs=False,
    num_coeffs=64,
):
    """
    Integrates the Mie parameters over an arbitrary particle size distribution, returning cross sections and phase matrices
      at a set of wavelengths.

    Note that the units of the input parameters are left arbitrary, but they must be consistent between each other,
    i.e., if wavelengths is specified in nm it is expected that the particle size distribution will be as a function
    of nm.

    Parameters
    ----------
    mie : sk.Mie
        Internal Mie algorithms to use
    prob_dist : rv_continuous
        Instance of a probability distribution to integrate over.  Distribution must be specified in the same units
        as wavelengths
    refrac_index_fn : fn
        Function taking in one argument (wavelength) and returning a complex number for refractive index at that
        wavelength.  Function argument should be in the same units as wavelengths.
    wavelengths : np.array
        Wavelengths to calculate for.  Units should be consistent with prob_dist
    num_angles: int, Optional
        Number of angles at which to evaluate p11, p12, p33, p34
    num_quad : int, Optional
        Number of quadrature points to use when integrating over particle size.  Default is 1024
    maxintquantile : float
        Used to determine the maximum radius to use in integration.  A value of 0.99 means that at least 99% of the
        probability distribution area multipied by radius**2 will be included within the integration bounds.  Default
        0.99999
    compute_coeffs : bool
        Option to also compute and return the greek coefficients. Default False.
    num_coeffes : int
        Optional parameter, maximum number of coefficients to return in the expansion. Default 64.
    Returns
    -------
    xr.Dataset
        Dataset containing the keys 'p11', 'p12', 'p33', 'p34', 'xs_total', 'xs_absorption', and 'xs_scattering'
        All scattering cross sections will be in units of wavelength**2. If compute_coeffs is True, will also
        contain keys 'lm_a1', 'lm_a2', 'lm_a3', 'lm_a4', 'lm_b1', and 'lm_b2'.
    """
    t_full = time.time()

    angles = np.linspace(0, 180, num_angles)

    # TODO FROM HERE
    norm = integrate.quad(
        lambda x: prob_dist.pdf(x) * x**2, 0, 1e25, points=(prob_dist.mean())
    )[0]

    def pdf(x):
        return prob_dist.pdf(x) * x**2 / norm

    # We have to determine the maximum R to integrate to, this is apparently very challenging, what we do is
    # calculate the CDF of repeatadly large values until it is greater than our threshold
    max_r = prob_dist.mean()
    while (
        integrate.quad(pdf, 0, max_r * 2, points=(prob_dist.mean()))[0]
        - integrate.quad(pdf, 0, max_r, points=(prob_dist.mean()))[0]
    ) > (1 - maxintquantile):
        max_r *= 2

    from scipy.special import roots_legendre

    x, w = roots_legendre(num_quad)
    x = 0.5 * (x + 1) * max_r
    w *= max_r / 2
    # TODO TO HERE, need to fix

    all_output = xr.Dataset(
        {
            "p11": (["wavelength", "angle"], np.zeros((len(wavelengths), len(angles)))),
            "p12": (["wavelength", "angle"], np.zeros((len(wavelengths), len(angles)))),
            "p33": (["wavelength", "angle"], np.zeros((len(wavelengths), len(angles)))),
            "p34": (["wavelength", "angle"], np.zeros((len(wavelengths), len(angles)))),
            "xs_total": (["wavelength"], np.zeros(len(wavelengths))),
            "xs_scattering": (["wavelength"], np.zeros(len(wavelengths))),
            "xs_absorption": (["wavelength"], np.zeros(len(wavelengths))),
        },
        coords={"wavelength": wavelengths, "angle": angles},
    )

    for idx, wavelength in enumerate(wavelengths):
        n = np.cdouble(refrac_index_fn(wavelength))

        size_param = 2 * np.pi * x / wavelength
        t = time.time()
        output = my_mie.calculate(size_param, n, np.cos(angles * np.pi / 180), False)
        logging.debug(f"Internal Mie Calculation {time.time() - t}")
        params = _integrable_parameters(output, prob_dist.pdf(x), x)

        # performing the integral
        all_output["p11"].values[idx, :] = np.sum(
            params["p11"] * np.transpose([w]), axis=0
        )
        all_output["p12"].values[idx, :] = np.sum(
            params["p12"] * np.transpose([w]), axis=0
        )
        all_output["p33"].values[idx, :] = np.sum(
            params["p33"] * np.transpose([w]), axis=0
        )
        all_output["p34"].values[idx, :] = np.sum(
            params["p34"] * np.transpose([w]), axis=0
        )
        all_output["xs_total"].values[idx] = np.sum(params["Cext"] * w)
        all_output["xs_scattering"].values[idx] = np.sum(params["Csca"] * w)
        all_output["xs_absorption"].values[idx] = np.sum(params["Cabs"] * w)
    _post_process(all_output, wavelengths)

    if compute_coeffs:
        # Note that P22 = P11 and P44 = p33
        lm_a1, lm_a2, lm_a3, lm_a4, lm_b1, lm_b2 = compute_greek_coefficients(
            p11=all_output["p11"].data,
            p12=all_output["p12"].data,
            p22=all_output["p11"].data,
            p33=all_output["p33"].data,
            p34=all_output["p34"].data,
            p44=all_output["p33"].data,
            angle_grid=angles,
            num_coeff=num_coeffs,
        )
        coeffs_output = xr.Dataset(
            {
                "lm_a1": (["wavelength", "legendre"], lm_a1),
                "lm_a2": (["wavelength", "legendre"], lm_a2),
                "lm_a3": (["wavelength", "legendre"], lm_a3),
                "lm_a4": (["wavelength", "legendre"], lm_a4),
                "lm_b1": (["wavelength", "legendre"], lm_b1),
                "lm_b2": (["wavelength", "legendre"], lm_b2),
            },
            coords={"wavelength": wavelengths, "legendre": np.arange(num_coeffs)},
        )
        all_output = all_output.merge(coeffs_output)

    logging.debug(f"Total Mie Calculation {time.time() - t_full}")
    return all_output


class ParticleSizeDistribution(abc.ABC):
    def __init__(self, identifier: str) -> None:
        """
        Abstract class to define particle size distributions that Mie parameters can be
        integrated over.  This class is a light wrapper on top of scipy.stats.rv_continuous
        which adds some additional information.

        Parameters
        ----------
        identifier : str
            A unique identifier for the distribution
        """
        self._identifier = identifier

    @abc.abstractmethod
    def distribution(self, **kwargs) -> rv_continuous:
        """
        Returns back the scipy object representing this distribution

        Returns
        -------
        rv_continuous
        """
        return self._distribution

    @property
    def identifier(self) -> str:
        """
        Get the unique identifier for this distribution

        Returns
        -------
        str
        """
        return self._identifier

    @abc.abstractmethod
    def args(self):
        """
        A list of arguments that are required to define this distribution when calling distribution
        """

    def freeze(self, **kwargs):
        """
        Freeze some of the arguments of this distribution. E.g. if `y` is an argument of this distrubtion, calling
        `freeze(y=5)` will return a new distribution that is the same as this one, but with `y` frozen to 5.

        Returns
        -------
        ParticleSizeDistribution
            A new distribution with some of args frozen
        """
        return FrozenDistribution(self, kwargs)


class LogNormalDistribution(ParticleSizeDistribution):
    def __init__(self) -> None:
        """
        A log normal particle size distribution, defined by two parameters, the median radius and mode width
        """
        super().__init__("lognormal")

    def distribution(self, **kwargs):
        return lognorm(np.log(kwargs["mode_width"]), scale=kwargs["median_radius"])

    def args(self):
        return ["median_radius", "mode_width"]


class FrozenDistribution(ParticleSizeDistribution):
    def __init__(
        self, base_distribution: ParticleSizeDistribution, frozen_parameters: dict
    ) -> None:
        """
        A particle size distribution that is frozen in time, useful for testing

        Parameters
        ----------
        base_distribution : ParticleSizeDistribution
            The distribution to freeze
        """
        identifier = f"frozen_{base_distribution.identifier}"
        for key, value in frozen_parameters.items():
            identifier += f"_{key}_{value}"

            if key not in base_distribution.args():
                msg = f"Frozen key {key} not in base distribution args"
                raise ValueError(msg)

        super().__init__(identifier)
        self._distribution = base_distribution

        self._frozen_parameters = frozen_parameters
        self._args = [
            arg for arg in base_distribution.args() if arg not in frozen_parameters
        ]

    def distribution(self, **kwargs):
        return self._distribution.distribution(**{**self._frozen_parameters, **kwargs})

    def args(self):
        return self._args



from __future__ import annotations

import numpy as np

from sasktran2.units import kelvin_to_celsius


def ciddor_index_of_refraction(
    temperatures_k: np.array,
    pressure_pa: np.array,
    specific_humidity: np.array,
    x_co2: np.array,
    wavelength_nm: float,
) -> np.array:
    """
    Calculates the index of refraction of air using the formula of Ciddor (1996).

    Parameters
    ----------
    temperatures_k : np.array
        Temperature in Kelvin
    pressure_pa : np.array
        Pressures in Pascals
    specific_humidity : np.array
        Specific humidity [kg/kg]
    x_co2 : np.array
        Molar fraction of CO2 in [ppm]
    wavelength_nm : float
        Wavelength in [nm]

    Returns
    -------
    np.array
        Array of index of refraction values at the given conditions
    """
    temperatures_c = kelvin_to_celsius(temperatures_k)
    # Convert specific humidity to VMR
    x_h2o = specific_humidity / (
        specific_humidity + (1 - specific_humidity) / (18.01528 / 28.9647)
    )  #  Molar mass of water vapor / Molar mass of air

    # Constants
    w0 = 295.235
    w1 = 2.6422
    w2 = -0.03238
    w3 = 0.004028

    k0 = 238.0185
    k1 = 5792105
    k2 = 57.362
    k3 = 167917

    a0 = 1.58123e-6
    a1 = -2.9331e-8
    a2 = 1.1043e-10

    b0 = 5.707e-6
    b1 = -2.051e-8

    c0 = 1.9898e-4
    c1 = -2.376e-6

    d = 1.83e-11
    e = -0.765e-8

    Pr1 = 101325
    TR1 = 288.15

    Za = 0.9995922115
    Pvs = 0.00985938

    R = 8.314472
    Mv = 0.018015

    S = 1 / ((wavelength_nm / 1e3) ** 2)  # Convert wavelength to micrometers

    ras = 1e-8 * (k1 / (k0 - S) + k3 / (k2 - S))

    rvs = 1.022e-8 * (w0 + w1 * S + w2 * S**2 + w3 * S**3)

    Ma = 0.0289635 + 1.2011e-8 * (x_co2 - 400)

    raxs = ras * (1 + 5.34e-7 * (x_co2 - 450))

    T = temperatures_c + 273.15

    Zm = (
        1
        - (pressure_pa / T)
        * (
            a0
            + a1 * temperatures_c
            + a2 * temperatures_c**2
            + (b0 + b1 * temperatures_c) * x_h2o
            + (c0 + c1 * temperatures_c) * x_h2o**2
        )
        + (pressure_pa / T) ** 2 * (d + e * x_h2o**2)
    )

    rho_axs = Pr1 * Ma / (Za * R * TR1)

    rho_v = x_h2o * pressure_pa * Mv / (Zm * R * T)

    rho_a = (1 - x_h2o) * pressure_pa * Ma / (Zm * R * T)

    return 1 + (rho_a / rho_axs) * raxs + (rho_v / Pvs) * rvs



from __future__ import annotations

import itertools
import os

_counter = itertools.count()


def parameterized(names, params):
    def decorator(func):
        func.param_names = names
        func.params = params
        return func

    return decorator


def requires_dask():
    try:
        import dask  # noqa: F401
    except ImportError:
        raise NotImplementedError()  # noqa: B904


def _skip_slow():
    """
    Use this function to skip slow or highly demanding tests.

    Use it as a `Class.setup` method or a `function.setup` attribute.

    Examples
    --------
    >>> from . import _skip_slow
    >>> def time_something_slow():
    ...     pass
    ...
    >>> time_something.setup = _skip_slow
    """
    if os.environ.get("ASV_SKIP_SLOW", "0") == "1":
        msg = "Skipping this test..."
        raise NotImplementedError(msg)



from __future__ import annotations

import numpy as np
from sasktran2.optical.rayleigh import rayleigh_cross_section_bates


def test_rayleigh_bates():
    _ = rayleigh_cross_section_bates(np.array([0.350]))



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from sasktran2.mie.refractive import RefractiveIndex

from sasktran2.atmosphere import Atmosphere

from ..base import Constituent
from . import (
    SnowKokhanovskyStokes_1,
    SnowKokhanovskyStokes_3,
    WavelengthInterpolatorMixin,
)


class SnowKokhanovsky(Constituent, WavelengthInterpolatorMixin):
    def __init__(
        self,
        L: np.array = 3600000,
        M: np.array = 5.5e-8,
        refractive_index_fn: RefractiveIndex = None,
        wavelengths_nm: np.array = None,
        out_of_bounds_mode="zero",
    ) -> None:
        """
        Parameters
        ----------
        L : np.array, optional
            Kokhanovsky L parameter, by default 3600000
        M : np.array, optional
            Kokhanovsky M parameter, by default 5.5e-8
        wavelengths_nm : np.array, optional
            Wavelengths in [nm] that the parameters L, M is specified at, by default None indicating that L and M are scalar
        out_of_bounds_mode : str, optional
            One of ["extend" or "zero"], "extend" will extend the last/first value if we are
            interpolating outside the grid. "zero" will set the albedo to 0 outside of the
            grid boundaries, by default "zero"
        """
        Constituent.__init__(self)
        WavelengthInterpolatorMixin.__init__(
            self,
            wavelengths_nm=wavelengths_nm,
            wavenumbers_cminv=None,
            out_of_bounds_mode=out_of_bounds_mode,
            param_length=len(np.atleast_1d(L)),
        )
        self._L = np.atleast_1d(L)
        self._M = np.atleast_1d(M)

        if refractive_index_fn is None:
            from sasktran2.mie.refractive import Ice

            self._refractive_index_fn = Ice()
        else:
            self._refractive_index_fn = refractive_index_fn

    @property
    def L(self) -> np.array:
        return self._L

    @L.setter
    def L(self, L: np.array):
        self._L = np.atleast_1d(L)

    @property
    def M(self) -> np.array:
        return self._M

    @M.setter
    def M(self, M: np.array):
        self._M = np.atleast_1d(M)

    def add_to_atmosphere(self, atmo: Atmosphere):
        if atmo.wavelengths_nm is None:
            msg = (
                "Atmosphere must have wavelengths defined before using SnowKokhonovsky"
            )
            raise ValueError(msg)

        atmo.surface.brdf = (
            SnowKokhanovskyStokes_1()
            if atmo.nstokes == 1
            else SnowKokhanovskyStokes_3()
        )

        interp_matrix = self._interpolating_matrix(atmo)

        # args(0) is (chi + M) * L / wavelength_nm
        # Where chi is the imaginary part of the ice refractive index
        chi = -self._refractive_index_fn.refractive_index(atmo.wavelengths_nm).imag

        M_interp = interp_matrix @ self._M
        L_interp = interp_matrix @ self._L

        atmo.surface.brdf_args[0, :] = (chi + M_interp) * L_interp / atmo.wavelengths_nm
        atmo.surface.d_brdf_args[0][0, :] = 1

    def register_derivative(self, atmo: Atmosphere, name: str):
        # Start by constructing the interpolation matrix
        interp_matrix = self._interpolating_matrix(atmo)

        derivs = {}

        chi = -self._refractive_index_fn.refractive_index(atmo.wavelengths_nm).imag

        # L Deriv factors are (chi + M_interp) / wavelength_nm
        L_factor = (chi + (interp_matrix @ self._M)) / atmo.wavelengths_nm

        deriv_mapping = atmo.surface.get_derivative_mapping(f"wf_{name}_L")
        deriv_mapping.d_brdf[:] += L_factor.reshape(-1, 1)
        deriv_mapping.interpolator = interp_matrix
        deriv_mapping.interp_dim = f"{name}_wavelength"

        deriv_mapping = atmo.surface.get_derivative_mapping(f"wf_{name}_M")

        # M Deriv factors are L / wavelength_nm
        M_factor = (interp_matrix @ self._L) / atmo.wavelengths_nm

        deriv_mapping.d_brdf[:] += M_factor.reshape(-1, 1)
        deriv_mapping.interpolator = interp_matrix
        deriv_mapping.interp_dim = f"{name}_wavelength"

        return derivs



from __future__ import annotations

import numpy as np
import pytest
import sasktran2 as sk
from sasktran2 import appconfig


def test_mie_construction():
    mie = sk.mie.LinearizedMie()
    size_param = np.array([3.0, 4.0, 5.0])
    cos_angles = np.linspace(-1, 1, 100)
    refractive_index = 1.5 + 0.0j

    output = mie.calculate(size_param, refractive_index, cos_angles, True)

    np.testing.assert_allclose(output.size_parameter, size_param)
    np.testing.assert_allclose(output.cos_angles, cos_angles)


def test_qext_q_Sca_s1_s2():
    mie = sk.mie.LinearizedMie()
    refractive_index = 0.75
    size_param = np.array(
        [0.101, 0.2]
    )  # not checking values for 0.2, just there to show that we can calculate with multiple size param and angles
    angles = np.arange(0, 181, 30)
    cos_angles = np.cos(angles * np.pi / 180)
    output = mie.calculate(size_param, refractive_index, cos_angles, True)

    np.testing.assert_allclose(output.values.Qext[0], 0.000008, atol=1e-06, rtol=0)
    np.testing.assert_allclose(output.values.Qsca[0], 0.000008, atol=1e-06, rtol=0)

    np.testing.assert_allclose(
        output.values.S1[0][0].real, 2.048754e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][0].imag, -1.756419e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][0].real, 2.048754e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][0].imag, -1.756419e-04, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][1].real, 2.048754e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][1].imag, -1.755965e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][1].real, 1.774273e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][1].imag, -1.520629e-04, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][2].real, 2.048753e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][2].imag, -1.754726e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][2].real, 1.024377e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][2].imag, -8.771198e-05, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][3].real, 2.048751e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][3].imag, -1.753033e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][3].real, 1.845057e-15, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][3].imag, -3.247270e-08, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][4].real, 2.048750e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][4].imag, -1.751341e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][4].real, -1.024375e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][4].imag, 8.759147e-05, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][5].real, 2.048749e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][5].imag, -1.750103e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][5].real, -1.774269e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][5].imag, 1.515715e-04, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][4].real, 2.048749e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][5].imag, -1.749650e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][6].real, -2.048749e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][6].imag, 1.749650e-04, atol=1e-06, rtol=0
    )


def test_mie_integration():
    mie = sk.mie.LinearizedMie()
    refractive_index = 1 - 0.1j
    size_param = np.array(
        [1, 2, 3]
    )  # not checking values for these, just there to show that we can calculate with multiple size param and angles
    angles = np.linspace(0, 180, 3)
    cos_angles = np.cos(angles * np.pi / 180)
    output = mie.calculate(size_param, refractive_index, cos_angles, True)
    np.testing.assert_allclose(
        output.values.Qext[0], 0.25925991210268584, atol=1e-06, rtol=0
    )
    from scipy.stats import rv_continuous

    class gaussian_gen(rv_continuous):
        "Gaussian distribution"

        def _pdf(self, x):
            mu = 500  # 500
            sig = 100  # 100
            return np.exp(-(((x - mu) / sig) ** 2) / 2.0) / np.sqrt(2.0 * np.pi) / sig

    prob_dist = gaussian_gen(name="gaussian")

    def refrac_index_fn(w):
        return 1.0 - 0.1j + 0 * w

    wavelengths = np.array([400, 500])
    my_mie = sk.mie.LinearizedMie()

    out = sk.mie.integrate_mie(
        my_mie,
        prob_dist,
        refrac_index_fn,
        wavelengths,
        num_angles=3,
        num_quad=20,
        compute_coeffs=True,
        num_coeffs=2,
    )
    np.testing.assert_allclose(out["p11"][0][0].data, 8.11553105e01, atol=1e-06, rtol=0)
    np.testing.assert_allclose(
        out["p11"][0][1].data, 1.16641980e-02, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        out["p11"][0][2].data, 5.64571171e-03, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(out["p12"][0][0].data, 0, atol=1e-06, rtol=0)
    np.testing.assert_allclose(out["p12"][0][1].data, 0.01117573, atol=1e-06, rtol=0)
    np.testing.assert_allclose(out["p12"][0][2].data, 0, atol=1e-06, rtol=0)
    np.testing.assert_allclose(out["p33"][0][0].data, 8.11553105e01, atol=1e-06, rtol=0)
    np.testing.assert_allclose(
        out["p33"][0][1].data, -6.33399061e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        out["p33"][0][2].data, -5.64571171e-03, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(out["p34"][0][0].data, 0, atol=1e-06, rtol=0)
    np.testing.assert_allclose(out["p34"][0][1].data, -0.00099857, atol=1e-06, rtol=0)
    np.testing.assert_allclose(out["p34"][0][2].data, 0, atol=1e-06, rtol=0)

    np.testing.assert_allclose(
        out["xs_total"][0].data, 1077477.99015462, atol=1e-03, rtol=0
    )
    np.testing.assert_allclose(
        out["xs_scattering"][0].data, 348484.73075379, atol=1e-03, rtol=0
    )
    np.testing.assert_allclose(
        out["xs_absorption"][0].data, 728993.25940083, atol=1e-03, rtol=0
    )


def test_mie_database():
    _ = pytest.importorskip("sasktran")

    db_root_legacy = appconfig.database_root().joinpath("legacy")
    db_root_current = appconfig.database_root().joinpath("current")

    refrac = sk.mie.refractive.H2SO4()
    distribution = sk.mie.distribution.LogNormalDistribution()
    mie_db = sk.database.MieDatabase(
        distribution,
        refrac,
        wavelengths_nm=np.arange(270, 1000, 50.0),
        median_radius=[100, 150, 200],
        mode_width=[1.5, 1.7],
        db_root=db_root_legacy,
        max_legendre_moments=100,
    )
    mie_db.load_ds()

    mie_db_new = sk.database.MieDatabase(
        distribution,
        refrac,
        wavelengths_nm=np.arange(270, 1000, 50.0),
        median_radius=[100, 150, 200],
        mode_width=[1.5, 1.7],
        backend="sasktran2",
        db_root=db_root_current,
        max_legendre_moments=100,
    )

    # testing select legendre moments for both to make sure we get reasonably close results
    tolerance = 1e-7
    np.testing.assert_allclose(
        mie_db._database["lm_a1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        mie_db_new._database["lm_a1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        mie_db_new._database["lm_a1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        mie_db_new._database["lm_a1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        mie_db_new._database["lm_a1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        atol=tolerance,
        rtol=0,
    )

    np.testing.assert_allclose(
        mie_db._database["lm_a2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        mie_db_new._database["lm_a2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        mie_db_new._database["lm_a2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        mie_db_new._database["lm_a2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        mie_db_new._database["lm_a2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        atol=tolerance,
        rtol=0,
    )

    np.testing.assert_allclose(
        mie_db._database["lm_a3"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        mie_db_new._database["lm_a3"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a3"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        mie_db_new._database["lm_a3"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a3"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        mie_db_new._database["lm_a3"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a3"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        mie_db_new._database["lm_a3"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        atol=tolerance,
        rtol=0,
    )

    np.testing.assert_allclose(
        mie_db._database["lm_a4"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        mie_db_new._database["lm_a4"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a4"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        mie_db_new._database["lm_a4"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a4"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        mie_db_new._database["lm_a4"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_a4"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        mie_db_new._database["lm_a4"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        atol=tolerance,
        rtol=0,
    )

    np.testing.assert_allclose(
        mie_db._database["lm_b1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        mie_db_new._database["lm_b1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_b1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        mie_db_new._database["lm_b1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_b1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        mie_db_new._database["lm_b1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_b1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        mie_db_new._database["lm_b1"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        atol=tolerance,
        rtol=0,
    )

    np.testing.assert_allclose(
        mie_db._database["lm_b2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        mie_db_new._database["lm_b2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=0)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_b2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        mie_db_new._database["lm_b2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=20)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_b2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        mie_db_new._database["lm_b2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=40)
        .data,
        atol=tolerance,
        rtol=0,
    )
    np.testing.assert_allclose(
        mie_db._database["lm_b2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        mie_db_new._database["lm_b2"]
        .isel(median_radius=0, mode_width=0, wavelength_nm=0, legendre=60)
        .data,
        atol=tolerance,
        rtol=0,
    )


def test_mie_db_freezing():
    refrac = sk.mie.refractive.H2SO4()
    # try freezing
    distribution = sk.mie.distribution.LogNormalDistribution().freeze(
        median_radius=160, mode_width=1.6
    )
    mie_db_f = sk.database.MieDatabase(
        distribution,
        refrac,
        wavelengths_nm=np.arange(270, 1000, 50.0),
    )

    mie_db_f.load_ds()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'sasktran2'
copyright = '2025, USask-ARG'
author = 'USask-ARG'

from importlib.metadata import version as get_version

release: str = get_version('sasktran2')
# for example take major/minor
version: str = ".".join(release.split('.')[:2])

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'myst_nb',
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    #"sphinx_copybutton",
    "sphinx_design",
    #"sphinx_examples",
    #"sphinx_tabs.tabs",
    #"sphinx_thebe",
    #"sphinx_togglebutton",
    #"sphinxcontrib.bibtex",
    #"sphinxext.opengraph",
    # For the kitchen sink
    "sphinx.ext.todo",
]

autosummary_generate = True

templates_path = ['_templates']
exclude_patterns = []

autodoc_docstring_signature = True
autodoc_default_options = {
    'members': True,
    'show-inheritance': True,
    'member_order': 'groupwise'
}
autoclass_content = 'both'



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ['_static']
html_css_files = ["locals.css"]

html_theme_options = {
    "github_url": "https://github.com/usask-arg/sasktran2",
    "repository_url": "https://github.com/usask-arg/sasktran2",
    "repository_branch": "main",
    "path_to_docs": "docs/sphinx/source",
    "use_repository_button": True,
    "use_edit_page_button": True,
    "use_issues_button": True,
    "logo": {
      "image_light": "_static/sasktran-dark.png",
      "image_dark": "_static/sasktran-light.png",
   }
}

nb_execution_timeout = 300


myst_enable_extensions = [
    "dollarmath",
    "amsmath",
    "deflist",
    # "html_admonition",
    # "html_image",
    "colon_fence",
    # "smartquotes",
    # "replacements",
    # "linkify",
    # "substitution",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "dask": ("https://docs.dask.org/en/latest", None),
    "sparse": ("https://sparse.pydata.org/en/latest/", None),
    "xarray-tutorial": ("https://tutorial.xarray.dev/", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None)
    # "opt_einsum": ("https://dgasmith.github.io/opt_einsum/", None),
}



from __future__ import annotations

import numpy as np

import sasktran2 as sk
from sasktran2.optical import pressure_temperature_to_numberdensity

from ..optical.base import OpticalProperty
from .base import Constituent


class CollisionInducedAbsorber(Constituent):
    def __init__(self, optical_property: OpticalProperty, name: str):
        """
        An implementation of collision-induced absorption for quantities with known constant mole
        fractions.

        Air number density is estimated through the ideal gas law. Currently the mole fraction is applied
        to the air number density, though in the future it should be applied to dry air only.

        This Constituent requires that the atmosphere object have `temperature_k`, `pressure_pa`, and
        `wavelength_nm` all defined inside the :py:class:`sasktran2.Atmosphere` object.

        Parameters
        ----------
        optical_property : OpticalProperty
            Optical property that computes the collision-induced cross section in [m^5].
        name : str
            Molecule names to load mole fractions.  Currently supported: O2O2
        """
        self._optical_property = optical_property
        if name.lower() == "o2o2":
            self._fraction_product = 0.20964**2
        else:
            msg = f"Unknown name '{name}'"
            raise ValueError(msg)

    def add_to_atmosphere(self, atmo: sk.Atmosphere):
        """
        Parameters
        ----------
        atmo : sk.Atmosphere

        :meta private:
        """
        if atmo.wavelengths_nm is None:
            msg = "It is required to give the Atmosphere object wavelengths to use the O2O2 constituent"
            raise ValueError(msg)

        if atmo.pressure_pa is None:
            msg = "It is required to set the pressure_pa property in the Atmosphere object to use the O2O2 Constituent"
            raise ValueError(msg)

        if atmo.temperature_k is None:
            msg = "It is required to set the temperature_k property in the Atmosphere object to use the O2O2 Constituent"
            raise ValueError(msg)

        # Get the number density from the atmosphere object at the grid points (m^-3)
        num_dens = pressure_temperature_to_numberdensity(
            atmo.pressure_pa, atmo.temperature_k
        )

        self._optical_quants = self._optical_property.atmosphere_quantities(atmo)

        atmo.storage.total_extinction += (  # m^-1
            self._optical_quants.extinction  # m^5 (cross section)
            * (self._fraction_product * num_dens**2)[:, np.newaxis]
        )

    def register_derivative(self, atmo: sk.Atmosphere, name: str):
        number_density, dN_dP, dN_dT = pressure_temperature_to_numberdensity(
            atmo.pressure_pa, atmo.temperature_k, include_derivatives=True
        )
        derivs = {}

        deriv_names = []
        d_vals = []
        if atmo.calculate_pressure_derivative:
            deriv_names.append("pressure_pa")
            d_vals.append(dN_dP)
        if atmo.calculate_temperature_derivative:
            deriv_names.append("temperature_k")
            d_vals.append(dN_dT)

        # Contributions from the change in number density via pressure/temperature
        for deriv_name, dN_dX in zip(deriv_names, d_vals, strict=False):
            dk_dX = (
                2
                * self._fraction_product
                * (number_density * dN_dX)[:, np.newaxis]
                * self._optical_quants.extinction
            )
            deriv_mapping = atmo.storage.get_derivative_mapping(
                f"wf_{name}_{deriv_name}"
            )
            deriv_mapping.d_extinction[:] += dk_dX
            deriv_mapping.d_ssa[:] += (
                dk_dX
                * (self._optical_quants.ssa - atmo.storage.ssa)
                / atmo.storage.total_extinction
            )
            deriv_mapping.interpolator = np.eye(len(number_density))
            deriv_mapping.interp_dim = "altitude"
            deriv_mapping.assign_name = f"wf_{deriv_name}"

        if len(deriv_names) > 0:
            optical_derivs = self._optical_property.optical_derivatives(atmo=atmo)

            for key, val in optical_derivs.items():
                deriv_mapping = atmo.storage.get_derivative_mapping(
                    f"wf_{name}_{deriv_name}_{key}_xs"
                )
                deriv_mapping.d_extinction[:] += val.d_extinction
                deriv_mapping.d_ssa[:] += (
                    val.d_extinction
                    * (self._optical_quants.ssa - atmo.storage.ssa)
                    / atmo.storage.total_extinction
                )
                deriv_mapping.interpolator = (
                    np.eye(len(number_density))
                    * (self._fraction_product * number_density**2)[np.newaxis, :]
                )
                deriv_mapping.interp_dim = "altitude"
                deriv_mapping.assign_name = f"wf_{deriv_name}"

        return derivs



from __future__ import annotations

import logging

import numpy as np

from sasktran2.atmosphere import Atmosphere
from sasktran2.database.hitran_line import HITRANLineDatabase
from sasktran2.optical.base import OpticalProperty, OpticalQuantities
from sasktran2.spectroscopy import voigt_broaden
from sasktran2.util import get_hapi


class HITRANAbsorber(OpticalProperty):
    def __init__(self, molecule: str, line_fn="voigt", **kwargs):
        """
        Absorption cross sections calculated using discrete absorption lines from the HITRAN database
        and broadened using the internal SASKTRAN2 Voigt line shape function.

        Notes:

        The HITRAN database is not included with SASKTRAN2 and must be downloaded separately. The hitran-api package
        is required.

        Broadening is calculated "online" rather than using pre-computed tables. This is slower but allows for
        more flexibility in the calculation.

        Derivatives with respect to pressure/temperature are currently not supported

        Parameters
        ----------
        molecule : str
            HITRAN Molecule identifier, e.g. "CO2", "H2O", "O3", "CH4"
        line_fn : str
            Line shape function to use, currently only "voigt" is supported
        kwargs : dict
            Additional keyword arguments to pass to the line broadening function
        """
        self._line_db = HITRANLineDatabase().load_ds(molecule)
        self._molecule = molecule
        self._kwargs = kwargs

        if line_fn != "voigt":
            msg = "Only Voigt line shape is supported"
            raise ValueError(msg)

        if self._molecule.lower() == "co2":
            # CO2 is bugged
            self._line_db.local_iso_id.values[
                self._line_db.local_iso_id.values == 0
            ] = 10

    def atmosphere_quantities(self, atmo: Atmosphere, **kwargs) -> OpticalQuantities:
        result = OpticalQuantities(
            extinction=np.zeros(
                (len(atmo.model_geometry.altitudes()), len(atmo.wavelengths_nm))
            ),
            ssa=np.zeros(
                (len(atmo.model_geometry.altitudes()), len(atmo.wavelengths_nm))
            ),
        )

        result.extinction[:] = self.cross_sections(
            atmo.wavelengths_nm,
            atmo.model_geometry.altitudes(),
            temperature_k=atmo.temperature_k,
            pressure_pa=atmo.pressure_pa,
            num_threads=atmo._config.num_threads,
        ).T
        return result

    def optical_derivatives(self, atmo: Atmosphere, **kwargs) -> dict:  # noqa: ARG002
        return {}

    def cross_sections(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> OpticalQuantities:
        if "temperature_k" not in kwargs:
            msg = "Temperature must be provided to calculate the cross sections"
            raise ValueError(msg)
        if "pressure_pa" not in kwargs:
            msg = "Pressure must be provided to calculate the cross sections"
            raise ValueError(msg)

        num_threads = kwargs.get("num_threads", 1)

        hapi = get_hapi()

        wavenumbers_cminv = 1e7 / wavelengths_nm
        temperature_k = kwargs["temperature_k"]
        pressure_pa = kwargs["pressure_pa"]

        iso_keys = np.unique(self._line_db.local_iso_id.to_numpy())

        partition_ratio = np.zeros((len(temperature_k), len(iso_keys)))
        molecular_mass = np.zeros(len(iso_keys))

        for key in iso_keys:
            vals = hapi.partitionSum(
                self._line_db["molec_id"].to_numpy()[0],
                key,
                [*list(temperature_k), 296],
            )
            partition_ratio[:, key - 1] = vals[:-1] / vals[-1]
            molecular_mass[key - 1] = hapi.molecularMass(
                self._line_db["molec_id"].to_numpy()[0], key
            )

        sidx = np.argsort(wavenumbers_cminv)

        result = np.zeros(
            (len(wavenumbers_cminv), (len(pressure_pa))),
            order="f",
        )

        logging.debug(f"Starting Broadening for {self._molecule}")
        voigt_broaden(
            self._line_db.nu.to_numpy(),
            self._line_db.sw.to_numpy(),
            self._line_db.elower.to_numpy(),
            self._line_db.gamma_air.to_numpy(),
            self._line_db.gamma_self.to_numpy(),
            self._line_db.delta_air.to_numpy(),
            self._line_db.n_air.to_numpy(),
            self._line_db.local_iso_id.to_numpy(),
            partition_ratio,
            molecular_mass,
            pressure_pa / 101325.0,
            np.ones(len(pressure_pa)) * 0,
            temperature_k,
            wavenumbers_cminv[sidx],
            result,
            num_threads=num_threads,
            **self._kwargs,
        )
        result[sidx, :] = result

        return result / 1e4

    def cross_section_derivatives(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> dict:
        return {}
